/// <reference path='../../../third_party/typings/es6-promise/es6-promise.d.ts' />
/// <reference path='../../../third_party/freedom-typings/freedom-common.d.ts' />
/// <reference path='../../../third_party/freedom-typings/freedom-module-env.d.ts' />
/// <reference path='../../../third_party/freedom-typings/tcp-socket.d.ts' />
var logging = require('../../../third_party/uproxy-lib/logging/logging');
var handler = require('../../../third_party/uproxy-lib/handler/queue');
var log = new logging.Log('tcp');
// Indicates how a socket (server or client) terminated.
(function (SocketCloseKind) {
    SocketCloseKind[SocketCloseKind["WE_CLOSED_IT"] = 0] = "WE_CLOSED_IT";
    SocketCloseKind[SocketCloseKind["REMOTELY_CLOSED"] = 1] = "REMOTELY_CLOSED";
    SocketCloseKind[SocketCloseKind["NEVER_CONNECTED"] = 2] = "NEVER_CONNECTED";
    SocketCloseKind[SocketCloseKind["UNKOWN"] = 3] = "UNKOWN";
})(exports.SocketCloseKind || (exports.SocketCloseKind = {}));
var SocketCloseKind = exports.SocketCloseKind;
// Maximum per-server number of TCP connections.
var DEFAULT_MAX_CONNECTIONS = 1048576;
// Public only for unit tests.
function endpointOfSocketInfo(info) {
    var retval = {};
    if (typeof info.localAddress == 'string' && typeof info.localPort == 'number') {
        retval.bound = {
            address: info.localAddress,
            port: info.localPort
        };
    }
    if (typeof info.peerAddress == 'string' && typeof info.peerPort == 'number') {
        retval.remote = {
            address: info.peerAddress,
            port: info.peerPort
        };
    }
    return retval;
}
exports.endpointOfSocketInfo = endpointOfSocketInfo;
// Closes a socket, along with its freedomjs interface object.
function destroyFreedomSocket_(socket) {
    // Note:
    //   freedom['core.tcpsocket'].close != freedom['core.tcpsocket']().close
    // The former destroys the freedom interface & communication channels.
    // The latter is a method on the constructed interface object that is on
    // the instance of the freedomjs TCP socket API.
    var destroy = function () {
        freedom['core.tcpsocket'].close(socket);
    };
    return socket.close().then(destroy, function (e) {
        destroy();
        return e;
    });
}
// Promise and handler queue-based TCP server with freedomjs sockets.
// TODO: protection against multiple calls to methods such as listen
var Server = (function () {
    function Server(endpoint_, maxConnections_) {
        var _this = this;
        if (maxConnections_ === void 0) { maxConnections_ = DEFAULT_MAX_CONNECTIONS; }
        this.endpoint_ = endpoint_;
        this.maxConnections_ = maxConnections_;
        // Active connections to the server.
        // TODO: index by connectionId rather than socketID
        this.connections_ = {};
        this.connectionsQueue = new handler.Queue();
        this.onceListening_ = new Promise(function (F, R) {
            _this.fulfillListening_ = F;
            _this.rejectListening_ = R;
        });
        this.onceShutdown_ = new Promise(function (F, R) {
            _this.fulfillShutdown_ = F;
        });
        // Invoked when the socket terminates.
        this.onDisconnectHandler_ = function (info) {
            log.debug('%1: disconnected: %2', _this.id_, JSON.stringify(info));
            if (info.errcode === 'SUCCESS') {
                _this.fulfillShutdown_(0 /* WE_CLOSED_IT */);
            }
            else {
                // TODO: investigate which other values occur
                _this.fulfillShutdown_(3 /* UNKOWN */);
            }
        };
        // Listens for connections, returning onceListening.
        // Should only be called once.
        this.listen = function () {
            _this.socket_.listen(_this.endpoint_.address, _this.endpoint_.port).then(_this.socket_.getInfo).then(function (info) {
                _this.endpoint_ = {
                    address: info.localAddress,
                    port: info.localPort
                };
                _this.fulfillListening_(_this.endpoint_);
            }).catch(function (e) {
                _this.rejectListening_(new Error('failed to listen: ' + e.message));
            });
            return _this.onceListening_;
        };
        // Invoked each time a new connection is established with the server.
        this.onConnectionHandler_ = function (acceptValue) {
            log.debug('%1: new connection', _this.id_);
            var socketId = acceptValue.socket;
            if (_this.connectionsCount() >= _this.maxConnections_) {
                log.warn('%1: hit maximum connections count, dropping new connection', _this.id_);
                destroyFreedomSocket_(freedom['core.tcpsocket'](socketId));
                return;
            }
            var connection = new Connection({
                existingSocketId: socketId
            });
            _this.connections_[socketId] = connection;
            var discard = function () {
                delete _this.connections_[socketId];
                log.debug('%1: discarded connection (%2 remaining)', _this.id_, _this.connectionsCount());
            };
            connection.onceClosed.then(discard, function (e) {
                log.error('%1: connection %2 rejected on close: %3', _this.id_, socketId, e.message);
                discard();
            });
            _this.connectionsQueue.handle(connection);
        };
        // Closes the server socket then closes all active connections.
        // Equivalent to calling stopListening followed by closeAll.
        this.shutdown = function () {
            log.debug('%1: shutdown', _this.id_);
            // This order is important: make sure no new connections happen while
            // we're trying to close all the connections.
            return _this.stopListening().then(_this.closeAll);
        };
        // Closes the server socket, preventing any new connections.
        // Does not affect active connections to the server.
        this.stopListening = function () {
            log.debug('%1: closing socket, no new connections will be accepted', _this.id_);
            return destroyFreedomSocket_(_this.socket_);
        };
        // Closes all active connections.
        this.closeAll = function () {
            log.debug('%1: closing all connections', _this.id_);
            var promises = [];
            for (var socketId in _this.connections_) {
                var connection = _this.connections_[socketId];
                promises.push(connection.close());
            }
            return Promise.all(promises).then(function (unused) {
            });
        };
        // Returns all active connections.
        this.connections = function () {
            var connections = [];
            for (var socketId in _this.connections_) {
                connections.push(_this.connections_[socketId]);
            }
            return connections;
        };
        // Returns the number of the active connections.
        this.connectionsCount = function () {
            return Object.keys(_this.connections_).length;
        };
        // Returns a promise which fulfills once the socket is accepting
        // connections and rejects if there is any error creating the socket
        // or listening for connections.
        this.onceListening = function () {
            return _this.onceListening_;
        };
        // Returns a promise which fulfills once the socket has stopped
        // accepting new connections, or the call to listen has failed.
        this.onceShutdown = function () {
            return _this.onceShutdown_;
        };
        this.toString = function () {
            return 'TCP server ' + _this.id_ + ': ' + JSON.stringify(_this.endpoint_) + ', ' + _this.connectionsCount() + ' connections: ' + _this.connections().join(', ');
        };
        this.id_ = 'S' + (Server.numCreations_++);
        this.onceListening_.catch(function (e) {
            _this.fulfillShutdown_(2 /* NEVER_CONNECTED */);
        });
        this.socket_ = freedom['core.tcpsocket']();
        this.socket_.on('onConnection', this.onConnectionHandler_);
        this.socket_.on('onDisconnect', this.onDisconnectHandler_);
    }
    // Count of sockets created to date.
    Server.numCreations_ = 0;
    return Server;
})();
exports.Server = Server;
// Tcp.Connection - Manages up a single TCP connection.
var Connection = (function () {
    // A TCP connection for a given socket.
    function Connection(connectionKind, startPaused_) {
        var _this = this;
        this.startPaused_ = startPaused_;
        // Receive returns a promise for exactly the next |ArrayBuffer| of data.
        this.receiveNext = function () {
            return new Promise(function (F, R) {
                _this.dataFromSocketQueue.setSyncNextHandler(F).catch(R);
            });
        };
        // Invoked when the socket is closed for any reason.
        // Fulfills onceClosed.
        this.onDisconnectHandler_ = function (info) {
            log.debug('%1: onDisconnect: %2', [
                _this.connectionId,
                JSON.stringify(info)
            ]);
            if (_this.state_ === 3 /* CLOSED */) {
                log.warn('%1: Got onDisconnect in closed state', [_this.connectionId]);
                return;
            }
            _this.state_ = 3 /* CLOSED */;
            _this.dataToSocketQueue.stopHandling();
            _this.dataToSocketQueue.clear();
            // CONSIDER: can this happen after a onceConnected promise rejection? if so,
            // do we want to preserve the SocketCloseKind.NEVER_CONNECTED result for
            // onceClosed?
            destroyFreedomSocket_(_this.connectionSocket_).then(function () {
                if (info.errcode === 'SUCCESS') {
                    _this.fulfillClosed_(0 /* WE_CLOSED_IT */);
                }
                else if (info.errcode === 'CONNECTION_CLOSED') {
                    _this.fulfillClosed_(1 /* REMOTELY_CLOSED */);
                }
                else {
                    _this.fulfillClosed_(3 /* UNKOWN */);
                }
            });
        };
        this.pause = function () {
            _this.connectionSocket_.pause();
        };
        this.resume = function () {
            _this.connectionSocket_.resume();
        };
        // This is called to close the underlying socket. This fulfills the
        // disconnect Promise `onceDisconnected`.
        this.close = function () {
            log.debug('%1: close', [_this.connectionId]);
            if (_this.state_ === 3 /* CLOSED */) {
                log.debug('%1: close called when already closed', [
                    _this.connectionId
                ]);
            }
            else {
                _this.connectionSocket_.close();
            }
            // The onDisconnect handler (which should only
            // be invoked once) actually stops handling, fulfills
            // onceClosed, etc.
            return _this.onceClosed;
        };
        // Boolean function to check if this connection is closed;
        this.isClosed = function () {
            return _this.state_ === 3 /* CLOSED */;
        };
        this.getState = function () {
            return _this.state_;
        };
        /**
         * Sends a message that is pre-formatted as an arrayBuffer.
         */
        this.send = function (msg) {
            return _this.dataToSocketQueue.handle(msg);
        };
        this.toString = function () {
            return 'Tcp.Connection(' + _this.connectionId + ':' + Connection.State[_this.state_] + ')';
        };
        this.connectionId = 'N' + Connection.globalConnectionId_++;
        this.dataFromSocketQueue = new handler.Queue();
        this.dataToSocketQueue = new handler.Queue();
        if (Object.keys(connectionKind).length !== 1) {
            //log.error(this.connectionId + ': Bad New Tcp Connection Kind:' +
            //       JSON.stringify(connectionKind));
            this.state_ = 0 /* ERROR */;
            this.onceConnected = Promise.reject(new Error(this.connectionId + 'Bad New Tcp Connection Kind:' + JSON.stringify(connectionKind)));
            this.onceClosed = Promise.resolve(2 /* NEVER_CONNECTED */);
            return;
        }
        if (connectionKind.existingSocketId) {
            // If we already have an open socket; i.e. from a previous tcp listen.
            // So we get a handler to the old freedom socket.
            this.connectionSocket_ = freedom['core.tcpsocket'](connectionKind.existingSocketId);
            this.onceConnected = this.connectionSocket_.getInfo().then(endpointOfSocketInfo);
            this.state_ = 2 /* CONNECTED */;
            this.connectionId = this.connectionId + '.A' + connectionKind.existingSocketId;
        }
        else if (connectionKind.endpoint) {
            // Create a new tcp socket to the given endpoint.
            this.connectionSocket_ = freedom['core.tcpsocket']();
            // We don't declare ourselves connected until we know the IP address to
            // which we have connected.  To speed this process up, we immediately
            // pause the socket as soon as it's connected, so that CPU time is not
            // wasted sending events that we can't pass on until getInfo returns.
            this.onceConnected = this.connectionSocket_.connect(connectionKind.endpoint.address, connectionKind.endpoint.port).then(this.pause).then(this.connectionSocket_.getInfo).then(function (info) {
                if (!_this.startPaused_) {
                    _this.resume();
                }
                return endpointOfSocketInfo(info);
            });
            this.state_ = 1 /* CONNECTING */;
            this.onceConnected.then(function () {
                // We need this guard because the getInfo call is async and a
                // close may happen affter the freedom socket connects and the
                // getInfo completes.
                if (_this.state_ !== 3 /* CLOSED */) {
                    _this.state_ = 2 /* CONNECTED */;
                }
            });
        }
        else {
            throw (new Error(this.connectionId + ': Should be impossible connectionKind' + JSON.stringify(connectionKind)));
        }
        // Use the dataFromSocketQueue handler for data from the socket.
        this.connectionSocket_.on('onData', function (readInfo) {
            _this.dataFromSocketQueue.handle(readInfo.data);
        });
        this.onceClosed = new Promise(function (F, R) {
            _this.fulfillClosed_ = F;
        });
        // Once we are connected, we start sending data to the underlying socket.
        // |dataToSocketQueue| allows a class using this connection to start
        // queuing data to be send to the socket.
        this.onceConnected.then(function () {
            _this.dataToSocketQueue.setHandler(_this.connectionSocket_.write);
        });
        this.onceConnected.catch(function (e) {
            _this.fulfillClosed_(2 /* NEVER_CONNECTED */);
        });
        this.connectionSocket_.on('onDisconnect', this.onDisconnectHandler_);
    }
    // Unique identifier for each connection.
    Connection.globalConnectionId_ = 0;
    return Connection;
})();
exports.Connection = Connection; // class Tcp.Connection
// Static stuff for the Connection class.
var Connection;
(function (Connection) {
    // Describes the state of a connection.
    (function (State) {
        State[State["ERROR"] = 0] = "ERROR";
        State[State["CONNECTING"] = 1] = "CONNECTING";
        State[State["CONNECTED"] = 2] = "CONNECTED";
        State[State["CLOSED"] = 3] = "CLOSED"; // Cannot change state.
    })(Connection.State || (Connection.State = {}));
    var State = Connection.State;
})(Connection = exports.Connection || (exports.Connection = {})); // module Connection
