// Server which handles SOCKS connections over WebRTC datachannels and send them
// out to the internet and sending back over WebRTC the responses.
var ipaddr = require('ipaddr.js');
var peerconnection = require('../../../third_party/uproxy-lib/webrtc/peerconnection');
var handler = require('../../../third_party/uproxy-lib/handler/queue');
var churn = require('../churn/churn');
var tcp = require('../net/tcp');
var socks = require('../socks-common/socks-headers');
var Pool = require('../pool/pool');
var logging = require('../../../third_party/uproxy-lib/logging/logging');
// module RtcToNet {
var log = new logging.Log('RtcToNet');
// The |RtcToNet| class holds a peer-connection and all its associated
// proxied connections.
// TODO: Extract common code for this and SocksToRtc:
//         https://github.com/uProxy/uproxy/issues/977
var RtcToNet = (function () {
    function RtcToNet() {
        var _this = this;
        // The two Queues below only count bytes transferred between the SOCKS
        // client and the remote host(s) the client wants to connect to. WebRTC
        // overhead (DTLS headers, ICE initiation, etc.) is not included (because
        // WebRTC does not provide easy access to that data) nor is SOCKS
        // protocol-related data (because it's sent via string messages).
        // All Sessions created in one instance of RtcToNet will share and
        // push numbers to the same queues (belonging to that instance of RtcToNet).
        // Queue of the number of bytes received from the peer. Handler is typically
        // defined in the class that creates an instance of RtcToNet.
        this.bytesReceivedFromPeer = new handler.Queue();
        // Queue of the number of bytes sent to the peer. Handler is typically
        // defined in the class that creates an instance of RtcToNet.
        this.bytesSentToPeer = new handler.Queue();
        this.onceStopping_ = new Promise(function (F, R) {
            _this.fulfillStopping_ = F;
        });
        // The connection to the peer that is acting as a proxy client. Once
        // assigned, is never un-assigned. Use in this class to tell if started.
        this.peerConnection_ = null;
        // The |sessions_| map goes from WebRTC data-channel labels to the Session.
        // Most of the wiring to manage this relationship happens via promises. We
        // need this only for data being received from a peer-connection data
        // channel get raised with data channel label. TODO:
        // https://github.com/uProxy/uproxy/issues/315 when closed allows
        // DataChannel and PeerConnection to be used directly and not via a freedom
        // interface. Then all work can be done by promise binding and this can be
        // removed.
        this.sessions_ = {};
        // As start() but handles creation of peerconnection.
        this.startFromConfig = function (proxyConfig, pcConfig, obfuscate) {
            if (pcConfig) {
                var pc = freedom['core.rtcpeerconnection'](pcConfig);
                return _this.start(proxyConfig, obfuscate ? new churn.Connection(pc, 'RtcToNet') : new peerconnection.PeerConnectionClass(pc));
            }
        };
        // Starts with the supplied peerconnection.
        // Returns this.onceReady.
        this.start = function (proxyConfig, peerconnection) {
            if (_this.peerConnection_) {
                throw new Error('already configured');
            }
            _this.peerConnection_ = peerconnection;
            _this.pool_ = new Pool(peerconnection, 'RtcToNet');
            _this.proxyConfig = proxyConfig;
            _this.signalsForPeer = _this.peerConnection_.signalForPeerQueue;
            _this.pool_.peerOpenedChannelQueue.setSyncHandler(_this.onPeerOpenedChannel_);
            // TODO: this.onceReady should reject if |this.onceStopping_|
            // fulfills first.  https://github.com/uProxy/uproxy/issues/760
            _this.onceReady = _this.peerConnection_.onceConnected.then(function () {
            });
            _this.onceReady.catch(_this.fulfillStopping_);
            _this.peerConnection_.onceDisconnected.then(function () {
                log.debug('peerconnection terminated');
            }, function (e) {
                log.error('peerconnection terminated with error: %1', [e.message]);
            }).then(_this.fulfillStopping_, _this.fulfillStopping_);
            _this.onceStopped = _this.onceStopping_.then(_this.stopResources_);
            // Uncomment this to see instrumentation data in the console.
            //this.onceReady.then(this.initiateSnapshotting);
            return _this.onceReady;
        };
        // Loops until onceStopped fulfills.
        this.initiateSnapshotting = function () {
            var loop = true;
            _this.onceStopped.then(function () {
                loop = false;
            });
            var writeSnapshot = function () {
                _this.getSnapshot().then(function (snapshot) {
                    log.info('snapshot: %1', JSON.stringify(snapshot));
                });
                if (loop) {
                    setTimeout(writeSnapshot, RtcToNet.SNAPSHOTTING_INTERVAL_MS);
                }
            };
            writeSnapshot();
        };
        // Snapshots the state of this RtcToNet instance.
        this.getSnapshot = function () {
            var promises = [];
            Object.keys(_this.sessions_).forEach(function (key) {
                promises.push(_this.sessions_[key].getSnapshot());
            });
            return Promise.all(promises).then(function (sessionSnapshots) {
                return {
                    sessions: sessionSnapshots
                };
            });
        };
        this.onPeerOpenedChannel_ = function (channel) {
            var channelLabel = channel.getLabel();
            log.info('associating session %1 with new datachannel', [channelLabel]);
            var session = new Session(channel, _this.proxyConfig, _this.bytesReceivedFromPeer, _this.bytesSentToPeer);
            _this.sessions_[channelLabel] = session;
            session.start().catch(function (e) {
                log.warn('session %1 failed to connect to remote endpoint: %2', [
                    channelLabel,
                    e.message
                ]);
            });
            var discard = function () {
                delete _this.sessions_[channelLabel];
                log.info('discarded session %1 (%2 remaining)', [
                    channelLabel,
                    Object.keys(_this.sessions_).length
                ]);
            };
            session.onceStopped().then(discard, function (e) {
                log.error('session %1 terminated with error: %2', [
                    channelLabel,
                    e.message
                ]);
                discard();
            });
        };
        // Initiates shutdown of the peerconnection.
        // Returns onceStopped.
        this.stop = function () {
            log.debug('stop requested');
            _this.fulfillStopping_();
            return _this.onceStopped;
        };
        // Shuts down the peerconnection, fulfilling once it has terminated.
        // Since its close() method should never throw, this should never reject.
        // TODO: close all sessions before fulfilling
        this.stopResources_ = function () {
            log.debug('freeing resources');
            // TODO(ldixon): explore why not not just return
            // this.peerConnection_.close(); call the PeerConnection's close and
            // return synchronously.
            return new Promise(function (F, R) {
                _this.peerConnection_.close();
                F();
            });
        };
        this.handleSignalFromPeer = function (message) {
            return _this.peerConnection_.handleSignalMessage(message);
        };
        this.toString = function () {
            var ret;
            var sessionsAsStrings = [];
            var label;
            for (label in _this.sessions_) {
                sessionsAsStrings.push(_this.sessions_[label].longId());
            }
            ret = JSON.stringify({ sessions_: sessionsAsStrings });
            return ret;
        };
    }
    // Time between outputting snapshots.
    RtcToNet.SNAPSHOTTING_INTERVAL_MS = 5000;
    return RtcToNet;
})();
exports.RtcToNet = RtcToNet; // class RtcToNet
// A Tcp connection and its data-channel on the peer connection.
//
// CONSIDER: when we have a lightweight webrtc provider, we can use the
// DataChannel class directly here instead of the awkward pairing of
// peerConnection with chanelLabel.
//
// CONSIDER: this and the socks-rtc session are similar: maybe abstract
// common parts into a super-class this inherits from?
var Session = (function () {
    // The supplied datachannel must already be successfully established.
    function Session(dataChannel_, proxyConfig_, bytesReceivedFromPeer_, bytesSentToPeer_) {
        var _this = this;
        this.dataChannel_ = dataChannel_;
        this.proxyConfig_ = proxyConfig_;
        this.bytesReceivedFromPeer_ = bytesReceivedFromPeer_;
        this.bytesSentToPeer_ = bytesSentToPeer_;
        this.onceStopping_ = new Promise(function (F, R) {
            _this.fulfillStopping_ = F;
        });
        this.onceStopped = function () {
            return _this.onceStopped_;
        };
        // Getters.
        this.channelLabel = function () {
            return _this.dataChannel_.getLabel();
        };
        this.socketSentBytes_ = 0;
        this.socketReceivedBytes_ = 0;
        this.channelSentBytes_ = 0;
        this.channelReceivedBytes_ = 0;
        // Returns onceReady.
        this.start = function () {
            _this.onceReady = _this.receiveEndpointFromPeer_().catch(function (e) {
                // TODO: Add a unit test for this case.
                _this.replyToPeer_(7 /* UNSUPPORTED_COMMAND */);
                return Promise.reject(e);
            }).then(_this.getTcpConnection_).then(function (tcpConnection) {
                _this.tcpConnection_ = tcpConnection;
                return _this.tcpConnection_.onceConnected.catch(function (e) {
                    log.info('%1: failed to connect to remote endpoint', [_this.longId()]);
                    _this.replyToPeer_(_this.getReplyFromError_(e));
                    return Promise.reject(new Error(e.errcode));
                });
            }).then(function (info) {
                log.info('%1: connected to remote endpoint', [_this.longId()]);
                log.debug('%1: bound address: %2', [_this.longId(), JSON.stringify(info.bound)]);
                var reply = _this.getReplyFromInfo_(info);
                _this.replyToPeer_(reply, info);
            });
            _this.onceReady.then(_this.linkSocketAndChannel_, _this.fulfillStopping_);
            // Shutdown once the data channel terminates.
            _this.dataChannel_.onceClosed.then(function () {
                if (_this.dataChannel_.dataFromPeerQueue.getLength() > 0) {
                    log.warn('%1: channel closed with %2 unprocessed incoming messages', _this.longId(), _this.dataChannel_.dataFromPeerQueue.getLength());
                }
                else {
                    log.info('%1: channel closed', _this.longId());
                }
                _this.fulfillStopping_();
            });
            _this.onceStopped_ = _this.onceStopping_.then(_this.stopResources_);
            return _this.onceReady;
        };
        // Initiates shutdown of the TCP connection and peerconnection.
        // Returns onceStopped.
        this.stop = function () {
            log.debug('%1: stop requested', [_this.longId()]);
            _this.fulfillStopping_();
            return _this.onceStopped_;
        };
        // Closes the TCP connection and datachannel if they haven't already
        // closed, fulfilling once both have closed. Since neither object's
        // close() methods should ever reject, this should never reject.
        this.stopResources_ = function () {
            log.debug('%1: freeing resources', [_this.longId()]);
            // DataChannel.close() returns void, implying that the shutdown is
            // effectively immediate.  However, we wrap it in a promise to ensure
            // that any exception is sent to the Promise.catch, rather than
            // propagating synchronously up the stack.
            var shutdownPromises = [
                new Promise(function (F, R) {
                    _this.dataChannel_.close();
                    F();
                })
            ];
            if (_this.tcpConnection_) {
                shutdownPromises.push(_this.tcpConnection_.close());
            }
            return Promise.all(shutdownPromises).then(function (discard) {
            });
        };
        // Fulfills with the endpoint requested by the SOCKS client.
        // Rejects if the received message is not for an endpoint
        // or if the received endpoint cannot be parsed.
        // TODO: needs tests (mocked by several tests)
        this.receiveEndpointFromPeer_ = function () {
            return new Promise(function (F, R) {
                _this.dataChannel_.dataFromPeerQueue.setSyncNextHandler(function (data) {
                    if (!data.str) {
                        R(new Error('received non-string data from peer: ' + JSON.stringify(data)));
                        return;
                    }
                    var request;
                    try {
                        request = JSON.parse(data.str);
                    }
                    catch (e) {
                        R(new Error('received malformed message during handshake: ' + data.str));
                        return;
                    }
                    if (!socks.isValidRequest(request)) {
                        R(new Error('received invalid request from peer: ' + JSON.stringify(data.str)));
                        return;
                    }
                    if (request.command != 1 /* TCP_CONNECT */) {
                        R(new Error('unexpected type for endpoint message'));
                        return;
                    }
                    // The domain name is very sensitive, so we keep it out of the
                    // info-level logs, which may be uploaded.
                    log.debug('%1: received endpoint from peer: %2', [
                        _this.longId(),
                        JSON.stringify(request.endpoint)
                    ]);
                    F(request.endpoint);
                    return;
                });
            });
        };
        this.getTcpConnection_ = function (endpoint) {
            if (ipaddr.isValid(endpoint.address) && !_this.isAllowedAddress_(endpoint.address)) {
                _this.replyToPeer_(2 /* NOT_ALLOWED */);
                throw new Error('tried to connect to disallowed address: ' + endpoint.address);
            }
            return new tcp.Connection({ endpoint: endpoint }, true);
        };
        // Fulfills once the connected endpoint has been returned to the SOCKS
        // client. Rejects if the endpoint cannot be sent to the SOCKS client.
        this.replyToPeer_ = function (reply, info) {
            var response = {
                reply: reply,
                endpoint: info ? info.bound : undefined
            };
            return _this.dataChannel_.send({
                str: JSON.stringify(response)
            }).then(function () {
                if (reply != 0 /* SUCCEEDED */) {
                    _this.stop();
                }
            });
        };
        this.getReplyFromInfo_ = function (info) {
            // TODO: This code should really return socks.Reply.NOT_ALLOWED,
            // but due to port-scanning concerns we return a generic error instead.
            // See https://github.com/uProxy/uproxy/issues/809
            return _this.isAllowedAddress_(info.remote.address) ? 0 /* SUCCEEDED */ : 1 /* FAILURE */;
        };
        this.getReplyFromError_ = function (e) {
            var reply = 1 /* FAILURE */;
            if (e.errcode == 'TIMED_OUT') {
                reply = 6 /* TTL_EXPIRED */;
            }
            else if (e.errcode == 'NETWORK_CHANGED') {
                reply = 3 /* NETWORK_UNREACHABLE */;
            }
            else if (e.errcode == 'CONNECTION_RESET' || e.errcode == 'CONNECTION_REFUSED') {
                // Due to port-scanning concerns, we return a generic error if the user
                // has blocked local network access and we are not sure if the requested
                // address might be on the local network.
                // See https://github.com/uProxy/uproxy/issues/809
                if (_this.proxyConfig_.allowNonUnicast) {
                    reply = 5 /* CONNECTION_REFUSED */;
                }
            }
            // TODO: report ConnectionInfo in cases where a port was bound.
            // Blocked by https://github.com/uProxy/uproxy/issues/803
            return reply;
        };
        // Sends a packet over the data channel.
        // Invoked when a packet is received over the TCP socket.
        this.sendOnChannel_ = function (data) {
            _this.socketReceivedBytes_ += data.byteLength;
            return _this.dataChannel_.send({ buffer: data });
        };
        // Sends a packet over the TCP socket.
        // Invoked when a packet is received over the data channel.
        this.sendOnSocket_ = function (data) {
            if (!data.buffer) {
                return Promise.reject(new Error('received non-buffer data from datachannel'));
            }
            _this.bytesReceivedFromPeer_.handle(data.buffer.byteLength);
            _this.channelReceivedBytes_ += data.buffer.byteLength;
            return _this.tcpConnection_.send(data.buffer);
        };
        // Configures forwarding of data from the TCP socket over the data channel
        // and vice versa. Should only be called once both socket and channel have
        // been successfully established.
        this.linkSocketAndChannel_ = function () {
            log.info('%1: linking socket and channel', _this.longId());
            var socketReader = function (data) {
                _this.sendOnChannel_(data).then(function () {
                    _this.bytesSentToPeer_.handle(data.byteLength);
                    _this.channelSentBytes_ += data.byteLength;
                }, function (e) {
                    log.error('%1: failed to send data on datachannel: %2', _this.longId(), e.message);
                });
            };
            _this.tcpConnection_.dataFromSocketQueue.setSyncHandler(socketReader);
            // Shutdown the session once the TCP connection terminates.
            // This should be safe now because
            // (1) this.tcpConnection_.dataFromPeerQueue has now been emptied into
            // this.dataChannel_.send() and (2) this.dataChannel_.close() should delay
            // closing until all pending messages have been sent.
            _this.tcpConnection_.onceClosed.then(function (kind) {
                log.info('%1: socket closed (%2)', _this.longId(), tcp.SocketCloseKind[kind]);
                _this.fulfillStopping_();
            });
            var channelReader = function (data) {
                _this.sendOnSocket_(data).then(function (writeInfo) {
                    _this.socketSentBytes_ += data.buffer.byteLength;
                }, function (e) {
                    // TODO: e is actually a freedom.Error (uproxy-lib 20+)
                    // errcode values are defined here:
                    //   https://github.com/freedomjs/freedom/blob/master/interface/core.tcpsocket.json
                    if (e.errcode === 'NOT_CONNECTED') {
                        // This can happen if, for example, there was still data to be
                        // read on the datachannel's queue when the socket closed.
                        log.warn('%1: tried to send data on closed socket: %2', [
                            _this.longId(),
                            e.errcode
                        ]);
                    }
                    else {
                        log.error('%1: failed to send data on socket: %2', [
                            _this.longId(),
                            e.errcode
                        ]);
                    }
                });
            };
            _this.dataChannel_.dataFromPeerQueue.setSyncHandler(channelReader);
            // The TCP connection starts in the paused state.  However, in extreme
            // cases, enough data can arrive before the pause takes effect to put
            // the data channel into overflow.  In that case, the socket will
            // eventually be resumed by the overflow listener below.
            if (!_this.dataChannel_.isInOverflow()) {
                _this.tcpConnection_.resume();
            }
            _this.dataChannel_.setOverflowListener(function (overflow) {
                if (_this.tcpConnection_.isClosed()) {
                    return;
                }
                if (overflow) {
                    _this.tcpConnection_.pause();
                    log.debug('%1: Hit overflow, pausing socket', _this.longId());
                }
                else {
                    _this.tcpConnection_.resume();
                    log.debug('%1: Exited  overflow, resuming socket', _this.longId());
                }
            });
        };
        this.isAllowedAddress_ = function (addressString) {
            // default is to disallow non-unicast addresses; i.e. only proxy for
            // public internet addresses.
            if (_this.proxyConfig_.allowNonUnicast) {
                return true;
            }
            try {
                var address = ipaddr.process(addressString);
                return address.range() == 'unicast';
            }
            catch (e) {
                // This likely indicates a malformed IP address, which will be logged by
                // the caller.
                return false;
            }
        };
        this.getSnapshot = function () {
            return _this.dataChannel_.getBrowserBufferedAmount().then(function (bufferedAmount) {
                var js_buffer = _this.dataChannel_.getJavascriptBufferedAmount();
                return {
                    name: _this.channelLabel(),
                    timestamp: performance.now(),
                    channel_sent: _this.channelSentBytes_,
                    channel_received: _this.channelReceivedBytes_,
                    channel_buffered: bufferedAmount,
                    channel_queue_size: _this.dataChannel_.dataFromPeerQueue.getLength(),
                    channel_queue_handling: _this.dataChannel_.dataFromPeerQueue.isHandling(),
                    channel_js_buffered: js_buffer,
                    socket_sent: _this.socketSentBytes_,
                    socket_received: _this.socketReceivedBytes_,
                    socket_queue_size: _this.tcpConnection_.dataFromSocketQueue.getLength(),
                    socket_queue_handling: _this.tcpConnection_.dataFromSocketQueue.isHandling()
                };
            });
        };
        this.longId = function () {
            var s = 'session ' + _this.channelLabel();
            if (_this.tcpConnection_) {
                s += ' (tcp-socket: ' + _this.tcpConnection_.connectionId + ' ' + (_this.tcpConnection_.isClosed() ? 'closed' : 'open') + ')';
            }
            return s;
        };
        // For logging/debugging.
        this.toString = function () {
            var tcpString = 'undefined';
            if (_this.tcpConnection_) {
                tcpString = _this.tcpConnection_.toString();
            }
            return JSON.stringify({
                channelLabel_: _this.channelLabel(),
                tcpConnection: tcpString
            });
        };
    }
    // Runs callback once the current event loop has run to completion.
    // Uses setTimeout in lieu of something like Node's process.nextTick:
    //   https://github.com/uProxy/uproxy/issues/967
    Session.nextTick_ = function (callback) {
        setTimeout(callback, 0);
    };
    return Session;
})();
exports.Session = Session; // Session
//}  // module RtcToNet
//export = RtcToNet;
