/// <reference path='../../../third_party/typings/es6-promise/es6-promise.d.ts' />
/// <reference path='../../../third_party/freedom-typings/freedom-module-env.d.ts' />
/// <reference path='../../../third_party/typings/jasmine/jasmine.d.ts' />
var freedomMocker = require('../../../third_party/uproxy-lib/freedom/mocks/mock-freedom-in-module-env');
freedom = freedomMocker.makeMockFreedomInModuleEnv();
var handler = require('../../../third_party/uproxy-lib/handler/queue');
var rtc_to_net = require('./rtc-to-net');
var tcp = require('../net/tcp');
var logging = require('../../../third_party/uproxy-lib/logging/logging');
var log = new logging.Log('socks-to-rtc spec');
var mockBoundEndpoint = {
    address: '127.0.0.1',
    port: 1234
};
var voidPromise = Promise.resolve();
var mockProxyConfig = {
    allowNonUnicast: false
};
var mockRemoteEndpoint = {
    // This address and port are both reserved for testing.
    address: '192.0.2.111',
    port: 1023
};
var mockConnectionInfo = {
    bound: mockBoundEndpoint,
    remote: mockRemoteEndpoint
};
// Neither fulfills nor rejects.
// Useful in a bunch of tests where a promise must be returned
// for chaining purposes.
var noopPromise = new Promise(function (F, R) {
});
describe('RtcToNet', function () {
    var server;
    var mockPeerconnection;
    beforeEach(function () {
        server = new rtc_to_net.RtcToNet();
        mockPeerconnection = {
            dataChannels: {},
            negotiateConnection: jasmine.createSpy('negotiateConnection'),
            onceConnecting: noopPromise,
            onceConnected: noopPromise,
            onceDisconnected: noopPromise,
            peerOpenedChannelQueue: new handler.Queue(),
            close: jasmine.createSpy('close')
        };
    });
    it('onceReady fulfills on peerconnection success', function (done) {
        mockPeerconnection.onceConnected = voidPromise;
        // We're not testing termination.
        server.start(mockProxyConfig, mockPeerconnection).then(done);
    });
    it('onceReady rejects on peerconnection setup failure', function (done) {
        mockPeerconnection.onceConnected = Promise.reject(new Error('failed to establish connection'));
        // We're not testing termination.
        server.start(mockProxyConfig, mockPeerconnection).catch(done);
    });
    it('onceStopped fulfills on peerconnection termination', function (done) {
        mockPeerconnection.onceConnected = voidPromise;
        mockPeerconnection.onceDisconnected = Promise.resolve();
        server.start(mockProxyConfig, mockPeerconnection).then(function () {
            return server.onceStopped;
        }).then(done);
    });
    it('onceStopped fulfills on call to stop', function (done) {
        mockPeerconnection.onceConnected = voidPromise;
        // Calling stop() alone should be sufficient to initiate shutdown.
        server.start(mockProxyConfig, mockPeerconnection).then(server.stop).then(function () {
            return server.onceStopped;
        }).then(done);
    });
});
describe("RtcToNet session", function () {
    var session;
    var mockTcpConnection;
    var mockDataChannel;
    var mockDataFromPeerQueue;
    var mockBytesReceived;
    var mockBytesSent;
    beforeEach(function () {
        mockTcpConnection = jasmine.createSpyObj('tcp connection', [
            'onceConnected',
            'onceClosed',
            'isClosed',
            'close',
            'send',
            'pause',
            'resume'
        ]);
        mockTcpConnection.send.and.returnValue(Promise.resolve({ bytesWritten: 1 }));
        mockTcpConnection.dataFromSocketQueue = new handler.Queue();
        mockDataFromPeerQueue = new handler.Queue();
        mockDataChannel = {
            close: jasmine.createSpy('close'),
            closeDataChannel: noopPromise,
            dataFromPeerQueue: mockDataFromPeerQueue,
            getLabel: jasmine.createSpy('getLabel'),
            onceClosed: noopPromise,
            send: jasmine.createSpy('send'),
            isInOverflow: jasmine.createSpy('isInOverflow').and.returnValue(false),
            setOverflowListener: jasmine.createSpy('setOverflowListener')
        };
        mockDataChannel.send.and.returnValue(voidPromise);
        mockBytesReceived = new handler.Queue();
        mockBytesSent = new handler.Queue();
        session = new rtc_to_net.Session(mockDataChannel, mockProxyConfig, mockBytesReceived, mockBytesSent);
    });
    it('onceReady fulfills with listening endpoint on successful negotiation', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        session.start().then(done);
    });
    it('onceReady rejects and onceStopped fulfills on unsuccessful endpoint negotiation', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.reject(new Error('bad format')));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        session.start().catch(session.onceStopped).then(done);
    });
    it('onceStopped fulfills on datachannel termination', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        mockDataChannel.onceClosed = Promise.resolve();
        session.start().then(session.onceStopped).then(done);
    });
    it('onceStopped fulfills on TCP connection termination', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = Promise.resolve(0 /* WE_CLOSED_IT */);
        session.start().then(session.onceStopped).then(done);
    });
    it('onceStopped fulfills on call to stop', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        session.start().then(session.stop).then(session.onceStopped).then(done);
    });
    it('bytes sent counter', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        var buffer = new Uint8Array([1, 2, 3]).buffer;
        session.start().then(function () {
            mockTcpConnection.dataFromSocketQueue.handle(buffer);
        });
        mockBytesSent.setSyncNextHandler(function (numBytes) {
            expect(numBytes).toEqual(buffer.byteLength);
            done();
        });
    });
    it('bytes received counter', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        var message = {
            buffer: new Uint8Array([1, 2, 3]).buffer
        };
        session.start().then(function () {
            mockDataFromPeerQueue.handle(message);
        });
        mockBytesReceived.setSyncNextHandler(function (numBytes) {
            expect(numBytes).toEqual(message.buffer.byteLength);
            done();
        });
    });
    it('channel queue drains before termination', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        // The data channel is closed before the session starts.
        mockDataChannel.onceClosed = voidPromise;
        var message = {
            buffer: new Uint8Array([1, 2, 3]).buffer
        };
        var onceMessageHandled = mockDataFromPeerQueue.handle(message);
        session.start().then(session.onceStopped).then(function () {
            return onceMessageHandled;
        }).then(function () {
            expect(mockDataChannel.dataFromPeerQueue.getLength()).toEqual(0);
            done();
        });
    });
    it('socket queue drains before termination', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        // The TCP connection is closed before the session starts.
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = Promise.resolve(0 /* WE_CLOSED_IT */);
        mockTcpConnection.isClosed.and.returnValue(true);
        var buffer = new Uint8Array([1, 2, 3]).buffer;
        var onceMessageHandled = mockTcpConnection.dataFromSocketQueue.handle(buffer);
        session.start().then(session.onceStopped).then(function () {
            return onceMessageHandled;
        }).then(function () {
            expect(mockTcpConnection.dataFromSocketQueue.getLength()).toEqual(0);
            done();
        });
    });
    it('backpressure', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        var overflowListener;
        mockDataChannel.setOverflowListener = function (listener) {
            overflowListener = listener;
        };
        var buffer = new Uint8Array([1, 2, 3]).buffer;
        // Messages received before start sit in the TCP receive queue.
        mockTcpConnection.dataFromSocketQueue.handle(buffer);
        expect(mockTcpConnection.dataFromSocketQueue.getLength()).toEqual(1);
        expect(mockDataChannel.send).not.toHaveBeenCalled();
        session.start().then(function () {
            // After start, the TCP queue should be drained into the datachannel.
            expect(mockTcpConnection.dataFromSocketQueue.getLength()).toEqual(0);
            expect(mockDataChannel.send).toHaveBeenCalled();
            // After draining the queue, the TCP connection should be resumed.
            expect(mockTcpConnection.pause).not.toHaveBeenCalled();
            expect(mockTcpConnection.resume).toHaveBeenCalled();
            // Enter overflow state.  This should trigger a call to pause.
            overflowListener(true);
            expect(mockTcpConnection.pause).toHaveBeenCalled();
            // In the paused state, messages are still forwarded
            mockTcpConnection.dataFromSocketQueue.handle(buffer);
            expect(mockTcpConnection.dataFromSocketQueue.getLength()).toEqual(0);
            expect(mockDataChannel.send.calls.count()).toEqual(2);
            // Exit overflow state.  This should trigger a call to resume.
            overflowListener(false);
            expect(mockTcpConnection.resume.calls.count()).toEqual(2);
            done();
        });
    });
    it('backpressure with early flood', function (done) {
        spyOn(session, 'receiveEndpointFromPeer_').and.returnValue(Promise.resolve(mockRemoteEndpoint));
        spyOn(session, 'replyToPeer_').and.returnValue(Promise.resolve());
        spyOn(session, 'getTcpConnection_').and.returnValue(Promise.resolve(mockTcpConnection));
        mockTcpConnection.onceConnected = Promise.resolve(mockConnectionInfo);
        mockTcpConnection.onceClosed = noopPromise;
        var overflowListener;
        mockDataChannel.setOverflowListener = function (listener) {
            overflowListener = listener;
        };
        mockDataChannel.isInOverflow = jasmine.createSpy('isInOverflow').and.returnValue(true);
        var buffer = new Uint8Array([1, 2, 3]).buffer;
        // Messages received before start sit in the TCP receive queue.
        mockTcpConnection.dataFromSocketQueue.handle(buffer);
        expect(mockTcpConnection.dataFromSocketQueue.getLength()).toEqual(1);
        expect(mockDataChannel.send).not.toHaveBeenCalled();
        session.start().then(function () {
            // After start, the TCP queue should be drained into the datachannel.
            expect(mockTcpConnection.dataFromSocketQueue.getLength()).toEqual(0);
            expect(mockDataChannel.send).toHaveBeenCalled();
            // If the initial queue is enough to trigger overflow, then the
            // socket should not be resumed.
            expect(mockTcpConnection.pause).not.toHaveBeenCalled();
            expect(mockTcpConnection.resume).not.toHaveBeenCalled();
            // Exit overflow state.  This should trigger a call to resume.
            overflowListener(false);
            expect(mockTcpConnection.resume).toHaveBeenCalled();
            done();
        });
    });
});
