/// <reference path='../../../third_party/freedom-typings/freedom-common.d.ts' />
/// <reference path='../../../third_party/freedom-typings/udp-socket.d.ts' />
/// <reference path='../../../third_party/typings/es6-promise/es6-promise.d.ts' />
var messages = require('../turn-frontend/messages');
var logging = require('../../../third_party/uproxy-lib/logging/logging');
var log = new logging.Log('TURN backend');
/**
 * Represents a client known to the server. One of these objects is created
 * in response to each ALLOCATE request.
 */
var Allocation = (function () {
    function Allocation() {
    }
    return Allocation;
})();
/**
 * Freedom module which handles relay sockets for the TURN server.
 */
var Backend = (function () {
    // TODO: define a type for event dispatcher in freedom-typescript-api
    function Backend(dispatchEvent_) {
        var _this = this;
        this.dispatchEvent_ = dispatchEvent_;
        /**
         * All clients currently known to the server, indexed by tag.
         * Note that this map is essentially the (extremely inaccurately) named
         * "5-tuple" introduced in section 2.2 of the TURN RFC:
         *   http://tools.ietf.org/html/rfc5766#section-2.2
         */
        this.allocations_ = {};
        this.handleIpc = function (data) {
            var request;
            try {
                request = messages.parseStunMessage(new Uint8Array(data));
            }
            catch (e) {
                return Promise.reject(new Error('failed to parse STUN message from IPC channel'));
            }
            // With which client is this message associated?
            var clientEndpoint;
            try {
                var ipcAttribute = messages.findFirstAttributeWithType(61183 /* IPC_TAG */, request.attributes);
                try {
                    clientEndpoint = messages.parseXorMappedAddressAttribute(ipcAttribute.value);
                }
                catch (e) {
                    return Promise.reject(new Error('could not parse address in IPC_TAG attribute: ' + e.message));
                }
            }
            catch (e) {
                return Promise.reject(new Error('message received on IPC channel without IPC_TAG attribute'));
            }
            var tag = clientEndpoint.address + ':' + clientEndpoint.port;
            if (request.method == 3 /* ALLOCATE */) {
                _this.makeAllocation_(clientEndpoint).then(function (allocation) {
                    allocation.socket.getInfo().then(function (socketInfo) {
                        _this.emitIpc_({
                            method: 3 /* ALLOCATE */,
                            clazz: 2 /* SUCCESS_RESPONSE */,
                            transactionId: request.transactionId,
                            attributes: [{
                                // Endpoint on which the new socket is listening.
                                // This is really the whole point of the thing.
                                type: 22 /* XOR_RELAYED_ADDRESS */,
                                value: messages.formatXorMappedAddressAttribute(socketInfo.localAddress, socketInfo.localPort)
                            }, {
                                // Endpoint from which the client appears to us.
                                // This is essentially a STUN response and is generally
                                // provided as a convenience to TURN clients.
                                type: 32 /* XOR_MAPPED_ADDRESS */,
                                value: messages.formatXorMappedAddressAttribute(clientEndpoint.address, clientEndpoint.port)
                            }, {
                                // Lifetime.
                                type: 13 /* LIFETIME */,
                                value: new Uint8Array([0x00, 0x00, 600 >> 8, 600 & 0xff]) // 600 = ten mins
                            }]
                        }, clientEndpoint);
                    });
                }, function (e) {
                    // Send error response (failed to make allocation).
                    _this.emitIpc_({
                        method: 3 /* ALLOCATE */,
                        clazz: 3 /* FAILURE_RESPONSE */,
                        transactionId: request.transactionId,
                        attributes: []
                    }, clientEndpoint);
                });
            }
            else if (request.method == 6 /* SEND */) {
                // Extract the destination address and payload.
                var destinationAttribute;
                var dataAttribute;
                try {
                    destinationAttribute = messages.findFirstAttributeWithType(18 /* XOR_PEER_ADDRESS */, request.attributes);
                    dataAttribute = messages.findFirstAttributeWithType(19 /* DATA */, request.attributes);
                }
                catch (e) {
                    return Promise.reject(new Error('no address or data attribute in SEND indication'));
                }
                var remoteEndpoint = messages.parseXorMappedAddressAttribute(destinationAttribute.value);
                var payload = Backend.bytesToArrayBuffer_(dataAttribute.value);
                if (!(tag in _this.allocations_)) {
                    return Promise.reject(new Error('received SEND indication for client without allocation'));
                }
                _this.allocations_[tag].then(function (allocation) {
                    allocation.socket.sendTo(payload, remoteEndpoint.address, remoteEndpoint.port);
                });
            }
            else {
                return Promise.reject(new Error('unsupported IPC method: ' + request.method));
            }
            return Promise.resolve();
        };
        /**
         * Emits a Freedom message which should be relayed to the remote side.
         * The message is a STUN message, as received from a TURN client but with
         * the addition of an IPC_TAG attribute identifying the TURN client.
         */
        this.emitIpc_ = function (stunMessage, clientEndpoint) {
            // Add an IPC_TAG attribute.
            stunMessage.attributes.push({
                type: 61183 /* IPC_TAG */,
                value: messages.formatXorMappedAddressAttribute(clientEndpoint.address, clientEndpoint.port)
            });
            _this.dispatchEvent_('ipc', {
                data: messages.formatStunMessage(stunMessage).buffer
            });
        };
        /** Promises to allocate a socket, wrapped in an Allocation. */
        this.makeAllocation_ = function (clientEndpoint) {
            var tag = clientEndpoint.address + ':' + clientEndpoint.port;
            if (tag in _this.allocations_) {
                return _this.allocations_[tag];
            }
            var socket = freedom['core.udpsocket']();
            var promise = socket.bind('127.0.0.1', 0).then(function (resultCode) {
                if (resultCode != 0) {
                    return Promise.reject(new Error('could not create socket -- error code ' + resultCode));
                }
                socket.getInfo().then(function (socketInfo) {
                    log.info('allocated socket for ' + tag + ' on ' + socketInfo.localAddress + ':' + socketInfo.localPort);
                });
                return Promise.resolve({
                    socket: socket
                });
            });
            socket.on('onData', function (recvFromInfo) {
                _this.emitIpc_({
                    method: 7 /* DATA */,
                    clazz: 4 /* INDICATION */,
                    transactionId: Backend.getRandomTransactionId_(),
                    attributes: [{
                        type: 18 /* XOR_PEER_ADDRESS */,
                        value: messages.formatXorMappedAddressAttribute(recvFromInfo.address, recvFromInfo.port)
                    }, {
                        type: 19 /* DATA */,
                        value: new Uint8Array(recvFromInfo.data)
                    }]
                }, clientEndpoint);
            });
            _this.allocations_[tag] = promise;
            return promise;
        };
        log.debug('TURN backend module created');
    }
    /**
     * Copies a Uint8Array into a new ArrayBuffer. Useful when the array
     * has been constructed from a subarray of the buffer, in which case
     * bytes.buffer is a much larger array than you are expecting.
     * TODO: be smarter about using slice in these instances
     */
    Backend.bytesToArrayBuffer_ = function (bytes) {
        var buffer = new ArrayBuffer(bytes.length);
        var view = new Uint8Array(buffer);
        for (var i = 0; i < bytes.length; i++) {
            view[i] = bytes[i];
        }
        return buffer;
    };
    Backend.getRandomTransactionId_ = function () {
        var bytes = new Uint8Array(20);
        for (var i = 0; i < 20; i++) {
            bytes[i] = Math.random() * 255;
        }
        return bytes;
    };
    return Backend;
})();
module.exports = Backend;
