/*
  For the RFC for socks, see:
    http://tools.ietf.org/html/rfc1928
*/
/// <reference path='../../../third_party/ipaddrjs/ipaddrjs.d.ts' />
var ipaddr = require('ipaddr.js');
// version 5 of socks
exports.VERSION5 = 0x05;
// AUTH - Authentication methods
(function (Auth) {
    Auth[Auth["NOAUTH"] = 0x00] = "NOAUTH";
    Auth[Auth["GSSAPI"] = 0x01] = "GSSAPI";
    Auth[Auth["USERPASS"] = 0x02] = "USERPASS";
    // X'03' to X'7F' IANA ASSIGNED
    // X'80' to X'FE' RESERVED FOR PRIVATE METHODS
    Auth[Auth["NONE"] = 0xFF] = "NONE"; // X'FF' NO ACCEPTABLE METHODS
})(exports.Auth || (exports.Auth = {}));
var Auth = exports.Auth;
// CMD - Commands
(function (Command) {
    Command[Command["TCP_CONNECT"] = 0x01] = "TCP_CONNECT";
    Command[Command["TCP_BIND"] = 0x02] = "TCP_BIND";
    Command[Command["UDP_ASSOCIATE"] = 0x03] = "UDP_ASSOCIATE"; // Connect to UDP association
})(exports.Command || (exports.Command = {}));
var Command = exports.Command;
// ATYP - address type of following address.
(function (AddressType) {
    AddressType[AddressType["IP_V4"] = 0x01] = "IP_V4";
    AddressType[AddressType["DNS"] = 0x03] = "DNS";
    AddressType[AddressType["IP_V6"] = 0x04] = "IP_V6"; // IP V6 Address
})(exports.AddressType || (exports.AddressType = {}));
var AddressType = exports.AddressType;
// REP - Reply Field
(function (Reply) {
    Reply[Reply["SUCCEEDED"] = 0x00] = "SUCCEEDED";
    Reply[Reply["FAILURE"] = 0x01] = "FAILURE";
    Reply[Reply["NOT_ALLOWED"] = 0x02] = "NOT_ALLOWED";
    Reply[Reply["NETWORK_UNREACHABLE"] = 0x03] = "NETWORK_UNREACHABLE";
    Reply[Reply["HOST_UNREACHABLE"] = 0x04] = "HOST_UNREACHABLE";
    Reply[Reply["CONNECTION_REFUSED"] = 0x05] = "CONNECTION_REFUSED";
    Reply[Reply["TTL_EXPIRED"] = 0x06] = "TTL_EXPIRED";
    Reply[Reply["UNSUPPORTED_COMMAND"] = 0x07] = "UNSUPPORTED_COMMAND";
    Reply[Reply["ADDRESS_TYPE"] = 0x08] = "ADDRESS_TYPE";
    Reply[Reply["RESERVED"] = 0x09] = "RESERVED"; // 0x09 - 0xFF unassigned
})(exports.Reply || (exports.Reply = {}));
var Reply = exports.Reply;
function isValidEndpoint(e) {
    if (typeof e != 'object') {
        return false;
    }
    if (typeof e.address != 'string') {
        return false;
    }
    if (typeof e.port != 'number' || e.port < 0 || e.port > 65535) {
        return false;
    }
    if (Object.keys(e).length > 2) {
        return false;
    }
    return true;
}
function isValidRequest(r) {
    if (typeof r != 'object') {
        return false;
    }
    if (typeof r.command != 'number' || typeof Command[r.command] != 'string') {
        return false;
    }
    if (!isValidEndpoint(r.endpoint)) {
        return false;
    }
    if (Object.keys(r).length > 2) {
        return false;
    }
    return true;
}
exports.isValidRequest = isValidRequest;
function isValidResponse(r) {
    if (typeof r != 'object') {
        return false;
    }
    if (typeof r.reply != 'number' || typeof Reply[r.reply] != 'string') {
        return false;
    }
    if (r.reply == 0 /* SUCCEEDED */ && !r.endpoint) {
        return false;
    }
    if (r.endpoint && !isValidEndpoint(r.endpoint)) {
        return false;
    }
    if (Object.keys(r).length > 2) {
        return false;
    }
    return true;
}
exports.isValidResponse = isValidResponse;
// Client to Server (Step 1)
//
// Examines the supplied session establishment bytes, throwing an
// error if the requested SOCKS version or METHOD is unsupported.
//
//   +----+----------+----------+
//   |VER | NMETHODS | METHODS  |
//   +----+----------+----------+
//   | 1  |    1     | 1 to 255 |
//   +----+----------+----------+
//
//
function interpretAuthHandshakeBuffer(buffer) {
    var handshakeBytes = new Uint8Array(buffer);
    // Only SOCKS Version 5 is supported.
    var socksVersion = handshakeBytes[0];
    if (socksVersion != exports.VERSION5) {
        throw new Error('unsupported SOCKS version: ' + socksVersion);
    }
    // Check AUTH methods on SOCKS handshake.
    // Get supported auth methods. Starts from 1, since 0 is already read.
    var authMethods = [];
    var numAuthMethods = handshakeBytes[1];
    for (var i = 0; i < numAuthMethods; i++) {
        authMethods.push(handshakeBytes[2 + i]);
    }
    // Make sure the client supports 'no authentication'.
    if (authMethods.indexOf(0 /* NOAUTH */) <= -1) {
        throw new Error('client requires authentication');
    }
    return authMethods;
}
exports.interpretAuthHandshakeBuffer = interpretAuthHandshakeBuffer;
function composeAuthHandshakeBuffer(auths) {
    if (auths.length == 0) {
        throw new Error('At least one authentication method must be specified.');
    }
    var handshakeBytes = new Uint8Array(auths.length + 2);
    handshakeBytes[0] = exports.VERSION5;
    handshakeBytes[1] = auths.length;
    handshakeBytes.set(auths, 2);
    return handshakeBytes.buffer;
}
exports.composeAuthHandshakeBuffer = composeAuthHandshakeBuffer;
// Server to Client (Step 2)
//
// Given an initial authentication query, compose a response with the support
// authentication types (none needed).
function composeAuthResponse(authType) {
    var buffer = new ArrayBuffer(2);
    var bytes = new Uint8Array(buffer);
    bytes[0] = exports.VERSION5;
    bytes[1] = authType;
    return buffer;
}
exports.composeAuthResponse = composeAuthResponse;
function interpretAuthResponse(buffer) {
    if (buffer.byteLength != 2) {
        throw new Error('Auth response must be exactly 2 bytes long');
    }
    var byteArray = new Uint8Array(buffer);
    // Only SOCKS Version 5 is supported.
    var socksVersion = byteArray[0];
    if (socksVersion != exports.VERSION5) {
        throw new Error('unsupported SOCKS version: ' + socksVersion);
    }
    return byteArray[1];
}
exports.interpretAuthResponse = interpretAuthResponse;
// Client to Server (Step 3-A)
//
// Interprets a SOCKS 5 request, which looks like this:
//
//   +----+-----+-------+------+----------+----------+
//   |VER | CMD |  RSV  | ATYP | DST.ADDR | DST.PORT |
//   +----+-----+-------+------+----------+----------+
//   | 1  |  1  | X'00' |  1   | Variable |    2     |
//   +----+-----+-------+------+----------+----------+
function interpretRequestBuffer(buffer) {
    return interpretRequest(new Uint8Array(buffer));
}
exports.interpretRequestBuffer = interpretRequestBuffer;
function interpretRequest(byteArray) {
    var version;
    var command;
    var destination;
    // Fail if the request is too short to be valid.
    if (byteArray.length < 9) {
        throw new Error('SOCKS request too short');
    }
    // Fail if client is not talking Socks version 5.
    var version = byteArray[0];
    if (version !== exports.VERSION5) {
        throw new Error('must be SOCKS5');
    }
    command = byteArray[1];
    // Fail unless we got a CONNECT or UDP_ASSOCIATE command.
    if (command != 1 /* TCP_CONNECT */ && command != 3 /* UDP_ASSOCIATE */) {
        throw new Error('unsupported SOCKS command (CMD): ' + command);
    }
    destination = interpretDestination(byteArray.subarray(3));
    var request = {
        command: command,
        endpoint: destination.endpoint
    };
    if (!isValidRequest(request)) {
        throw new Error('Constructed invalid request object: ' + JSON.stringify(request));
    }
    return request;
}
exports.interpretRequest = interpretRequest;
function composeRequestBuffer(request) {
    var destination = makeDestinationFromEndpoint(request.endpoint);
    // The header is 3 bytes
    var byteArray = new Uint8Array(3 + destination.addressByteLength);
    byteArray[0] = exports.VERSION5;
    byteArray[1] = request.command;
    byteArray[2] = 0; // reserved
    byteArray.set(composeDestination(destination), 3);
    return byteArray.buffer;
}
exports.composeRequestBuffer = composeRequestBuffer;
// Client to Server (Step 3-B)
//
// Interprets a SOCKS5 UDP request, returning the UInt8Array view of the
// sub-section for the DATA part of the request. The Request looks like this:
//
//   +----+------+------+----------+----------+----------+
//   |RSV | FRAG | ATYP | DST.ADDR | DST.PORT |   DATA   |
//   +----+------+------+----------+----------+----------+
//   | 2  |  1   |  1   | Variable |    2     | Variable |
//   +----+------+------+----------+----------+----------+
function interpretUdpMessage(byteArray) {
    // Fail if the request is too short to be valid.
    if (byteArray.length < 10) {
        throw new Error('UDP request too short');
    }
    var destination = interpretDestination(byteArray.subarray(3));
    var udpMessage = {
        frag: byteArray[2],
        destination: destination,
        data: byteArray.subarray(3 + destination.addressByteLength)
    };
    // Fail if client is requesting fragmentation.
    if (udpMessage.frag !== 0) {
        throw new Error('fragmentation not supported');
    }
    return udpMessage;
}
exports.interpretUdpMessage = interpretUdpMessage;
// Interprets this sub-structure, found within both "regular" SOCKS requests
// and UDP requests:
//
//   +------+----------+----------+
//   | ATYP | DST.ADDR | DST.PORT |
//   +------+----------+----------+
//   |  1   | Variable |    2     |
//   +------+----------+----------+
//
// The length of DST.ADDR varies according to the type found (IPv4, IPv6,
// host name, etc).
//
// Returns a Destination which captures this in a more convenient form.
function interpretDestination(byteArray) {
    var portOffset;
    var addressType;
    var addressSize;
    var address;
    var port;
    addressType = byteArray[0];
    if (1 /* IP_V4 */ == addressType) {
        addressSize = 4;
        address = interpretIpv4Address(byteArray.subarray(1, 1 + addressSize));
        portOffset = addressSize + 1;
    }
    else if (3 /* DNS */ == addressType) {
        addressSize = byteArray[1];
        address = '';
        for (var i = 0; i < addressSize; ++i) {
            address += String.fromCharCode(byteArray[2 + i]);
        }
        portOffset = addressSize + 2;
    }
    else if (4 /* IP_V6 */ == addressType) {
        addressSize = 16;
        address = interpretIpv6Address(byteArray.subarray(1, 1 + addressSize));
        portOffset = addressSize + 1;
    }
    else {
        throw new Error('Unsupported SOCKS address type: ' + addressType);
    }
    // Parse the port.
    port = byteArray[portOffset] << 8 | byteArray[portOffset + 1];
    return {
        addressType: addressType,
        endpoint: { address: address, port: port },
        addressByteLength: portOffset + 2
    };
}
exports.interpretDestination = interpretDestination;
function interpretIpv4Address(byteArray) {
    if (byteArray.length != 4) {
        throw new Error('IPv4 addresses must be exactly 4 bytes long');
    }
    var ipAddress = new ipaddr.IPv4(Array.prototype.slice.call(byteArray));
    return ipAddress.toString();
}
// Heler function for parsing an IPv6 address from an Uint8Array portion of
// a socks address in an arraybuffer.
function interpretIpv6Address(byteArray) {
    if (byteArray.length != 16) {
        throw new Error('IPv6 addresses must be exactly 16 bytes long');
    }
    // |byteArray| contains big-endian shorts, but Uint16Array would read it
    // as little-endian on most platforms, so we have to read it manually.
    var parts = [];
    for (var i = 0; i < 16; i += 2) {
        parts.push(byteArray[i] << 8 | byteArray[i + 1]);
    }
    var ipAddress = new ipaddr.IPv6(parts);
    return ipAddress.toString();
}
exports.interpretIpv6Address = interpretIpv6Address;
function composeDestination(destination) {
    var endpoint = destination.endpoint;
    var address = new Uint8Array(destination.addressByteLength);
    address[0] = destination.addressType;
    var addressSize;
    switch (destination.addressType) {
        case 1 /* IP_V4 */:
            addressSize = 4;
            var ipv4 = ipaddr.IPv4.parse(endpoint.address);
            address.set(ipv4.octets, 1);
            break;
        case 3 /* DNS */:
            addressSize = endpoint.address.length + 1;
            address[1] = endpoint.address.length;
            for (var i = 0; i < endpoint.address.length; ++i) {
                address[i + 2] = endpoint.address.charCodeAt(i);
            }
            break;
        case 4 /* IP_V6 */:
            addressSize = 16;
            var ipv6 = ipaddr.IPv6.parse(endpoint.address);
            address.set(ipv6.toByteArray(), 1);
            break;
        default:
            throw new Error('Unsupported SOCKS address type: ' + destination.addressType);
    }
    var portOffset = addressSize + 1;
    address[portOffset] = endpoint.port >> 8;
    address[portOffset + 1] = endpoint.port & 0xFF;
    return address;
}
exports.composeDestination = composeDestination;
function makeDestinationFromEndpoint(endpoint) {
    var type;
    var byteLength;
    if (ipaddr.IPv4.isValid(endpoint.address)) {
        type = 1 /* IP_V4 */;
        byteLength = 7; // 1 (type) + 4 (address) + 2 (port)
    }
    else if (ipaddr.IPv6.isValid(endpoint.address)) {
        type = 4 /* IP_V6 */;
        byteLength = 19; // 1 (type) + 16 (address) + 2 (port)
    }
    else {
        // TODO: Fail if the string is not a valid DNS name.
        type = 3 /* DNS */;
        // 4 = 1 (type) + 1 (length) + 2 (port)
        byteLength = endpoint.address.length + 4;
    }
    return {
        addressType: type,
        endpoint: endpoint,
        addressByteLength: byteLength
    };
}
// Server to Client (Step 4-A)
//
// TODO: support failure (https://github.com/uProxy/uproxy/issues/321)
//
// Given a destination reached, compose a response.
function composeResponseBuffer(response) {
    var fakeEndpoint = {
        address: '0.0.0.0',
        port: 0
    };
    var endpoint = response.endpoint || fakeEndpoint;
    var destination = makeDestinationFromEndpoint(endpoint);
    var destinationArray = composeDestination(destination);
    var bytes = new Uint8Array(destinationArray.length + 3);
    bytes[0] = exports.VERSION5;
    bytes[1] = response.reply;
    bytes[2] = 0x00;
    bytes.set(destinationArray, 3);
    return bytes.buffer;
}
exports.composeResponseBuffer = composeResponseBuffer;
function interpretResponseBuffer(buffer) {
    var bytes = new Uint8Array(buffer);
    // Only SOCKS Version 5 is supported.
    var socksVersion = bytes[0];
    if (socksVersion != exports.VERSION5) {
        throw new Error('unsupported SOCKS version: ' + socksVersion);
    }
    var reply = bytes[1];
    var destination = interpretDestination(bytes.subarray(3));
    var response = {
        reply: reply,
        endpoint: destination.endpoint
    };
    if (!isValidResponse(response)) {
        throw new Error('Constructed invalid response object: ' + JSON.stringify(response));
    }
    return response;
}
exports.interpretResponseBuffer = interpretResponseBuffer;
