/// <reference path='../../../third_party/typings/jasmine/jasmine.d.ts' />
var Socks = require('./socks-headers');
// TODO: add tests for IPv6 address parsing
describe("socks", function () {
    // A valid SOCKS5/IPV4 request.
    var ipv4RequestArray;
    // A valid SOCKS5/UDP request.
    var udpMessageArray;
    beforeEach(function () {
        ipv4RequestArray = new Uint8Array([
            Socks.VERSION5,
            1 /* TCP_CONNECT */,
            0,
            1 /* IP_V4 */,
            192,
            168,
            1,
            1,
            1200 >> 8,
            1200 & 0xFF
        ]); // port: 1200
        udpMessageArray = new Uint8Array([
            0,
            0,
            0,
            1 /* IP_V4 */,
            192,
            168,
            1,
            1,
            1200 >> 8,
            1200 & 0xFF,
            11,
            12
        ]); // datagram (byte 2/2)
    });
    it('roundtrip auth request', function () {
        var auths = [
            1 /* GSSAPI */,
            0 /* NOAUTH */,
            255 /* NONE */,
            2 /* USERPASS */
        ];
        var buffer = Socks.composeAuthHandshakeBuffer(auths);
        var authsAgain = Socks.interpretAuthHandshakeBuffer(buffer);
        expect(authsAgain).toEqual(auths);
    });
    it('roundtrip auth response', function () {
        var auth = 2 /* USERPASS */;
        var buffer = Socks.composeAuthResponse(auth);
        var authAgain = Socks.interpretAuthResponse(buffer);
        expect(authAgain).toEqual(auth);
    });
    it('reject wrongly sized requests', function () {
        expect(function () {
            Socks.interpretRequestBuffer(new ArrayBuffer(8));
        }).toThrow();
    });
    it('compose ipv4 tcp request', function () {
        var request = {
            command: 1 /* TCP_CONNECT */,
            endpoint: {
                address: '192.168.1.1',
                port: 1200
            }
        };
        var requestBuffer = Socks.composeRequestBuffer(request);
        var requestArray = new Uint8Array(requestBuffer);
        expect(requestArray).toEqual(ipv4RequestArray);
    });
    it('parse ipv4 request', function () {
        var result = Socks.interpretRequest(ipv4RequestArray);
        expect(result.command).toEqual(1 /* TCP_CONNECT */);
        expect(result.endpoint.address).toEqual('192.168.1.1');
        expect(result.endpoint.port).toEqual(1200);
    });
    it('roundtrip ipv6 tcp request', function () {
        var request = {
            command: 1 /* TCP_CONNECT */,
            endpoint: {
                address: '2620::1003:1003:a84f:9831:df45:5420',
                port: 1200
            }
        };
        var requestBuffer = Socks.composeRequestBuffer(request);
        var requestArray = new Uint8Array(requestBuffer);
        expect(requestArray[3]).toEqual(4 /* IP_V6 */);
        var requestAgain = Socks.interpretRequestBuffer(requestBuffer);
        expect(requestAgain).toEqual(request);
    });
    it('roundtrip DNS tcp request', function () {
        var request = {
            command: 1 /* TCP_CONNECT */,
            endpoint: {
                address: 'www.example.com',
                port: 1200
            }
        };
        var requestBuffer = Socks.composeRequestBuffer(request);
        var requestArray = new Uint8Array(requestBuffer);
        expect(requestArray[3]).toEqual(3 /* DNS */);
        var requestAgain = Socks.interpretRequestBuffer(requestBuffer);
        expect(requestAgain).toEqual(request);
    });
    it('roundtrip IPv4 response', function () {
        var response = {
            reply: 0 /* SUCCEEDED */,
            endpoint: {
                address: '255.0.1.77',
                port: 65535
            }
        };
        var responseBuffer = Socks.composeResponseBuffer(response);
        var responseArray = new Uint8Array(responseBuffer);
        expect(responseArray[3]).toEqual(1 /* IP_V4 */);
        var responseAgain = Socks.interpretResponseBuffer(responseBuffer);
        expect(responseAgain).toEqual(response);
    });
    it('roundtrip IPv6 request response', function () {
        var response = {
            reply: 1 /* FAILURE */,
            endpoint: {
                address: '2620::1003:1003:a84f:9831:df45:5420',
                port: 40000
            }
        };
        var responseBuffer = Socks.composeResponseBuffer(response);
        var responseArray = new Uint8Array(responseBuffer);
        expect(responseArray[3]).toEqual(4 /* IP_V6 */);
        var responseAgain = Socks.interpretResponseBuffer(responseBuffer);
        expect(responseAgain).toEqual(response);
    });
    it('roundtrip DNS request response', function () {
        var response = {
            reply: 2 /* NOT_ALLOWED */,
            endpoint: {
                address: 'www.subdomain.example.com',
                port: 45654
            }
        };
        var responseBuffer = Socks.composeResponseBuffer(response);
        var responseArray = new Uint8Array(responseBuffer);
        expect(responseArray[3]).toEqual(3 /* DNS */);
        var responseAgain = Socks.interpretResponseBuffer(responseBuffer);
        expect(responseAgain).toEqual(response);
    });
    it('wrong socks version', function () {
        ipv4RequestArray[0] = 4;
        expect(function () {
            Socks.interpretRequest(ipv4RequestArray);
        }).toThrow();
    });
    it('unsupported command', function () {
        ipv4RequestArray[1] = 2 /* TCP_BIND */;
        expect(function () {
            Socks.interpretRequest(ipv4RequestArray);
        }).toThrow();
    });
    it('parse destination', function () {
        var destination = Socks.interpretDestination(ipv4RequestArray.subarray(3));
        expect(destination.addressByteLength).toEqual(7);
        expect(destination.addressType).toEqual(1 /* IP_V4 */);
        expect(destination.endpoint.address).toEqual('192.168.1.1');
        expect(destination.endpoint.port).toEqual(1200);
    });
    it('parse udp request', function () {
        var udpMessage = Socks.interpretUdpMessage(udpMessageArray);
        expect(udpMessage.frag).toEqual(0);
        expect(udpMessage.destination.addressType).toEqual(1 /* IP_V4 */);
        expect(udpMessage.destination.endpoint.address).toEqual('192.168.1.1');
        expect(udpMessage.destination.endpoint.port).toEqual(1200);
        expect(udpMessage.data.byteLength).toEqual(2);
        expect(udpMessage.data[0]).toEqual(11);
        expect(udpMessage.data[1]).toEqual(12);
    });
    it('reject wrongly sized udp requests', function () {
        expect(function () {
            Socks.interpretUdpMessage(new Uint8Array(new ArrayBuffer(9)));
        }).toThrow();
    });
    it('reject fragmentation requests', function () {
        udpMessageArray[2] = 7;
        expect(function () {
            Socks.interpretUdpMessage(udpMessageArray);
        }).toThrow();
    });
});
