/// <reference path='../../../third_party/typings/es6-promise/es6-promise.d.ts' />
/// <reference path='../../../third_party/freedom-typings/freedom-common.d.ts' />
/// <reference path='../../../third_party/freedom-typings/udp-socket.d.ts' />
/// <reference path='../../../third_party/ipaddrjs/ipaddrjs.d.ts' />
var _this = this;
var handler = require('../../../third_party/uproxy-lib/handler/queue');
var ipaddr = require('ipaddr.js');
var logging = require('../../../third_party/uproxy-lib/logging/logging');
var peerconnection = require('../../../third_party/uproxy-lib/webrtc/peerconnection');
var signals = require('../../../third_party/uproxy-lib/webrtc/signals');
var log = new logging.Log('churn');
exports.filterCandidatesFromSdp = function (sdp) {
    return sdp.split('\n').filter(function (s) {
        return s.indexOf('a=candidate') != 0;
    }).join('\n');
};
var splitCandidateLine_ = function (candidate) {
    var lines = candidate.split(' ');
    if (lines.length < 8 || lines[6] != 'typ') {
        throw new Error('cannot parse candidate line: ' + candidate);
    }
    return lines;
};
var splitHostCandidateLine_ = function (candidate) {
    var lines = splitCandidateLine_(candidate);
    var typ = lines[7];
    if (typ != 'host') {
        throw new Error('not a host candidate line: ' + candidate);
    }
    return lines;
};
exports.extractEndpointFromCandidateLine = function (candidate) {
    var lines = splitHostCandidateLine_(candidate);
    var address = lines[4];
    var port = parseInt(lines[5]);
    if (port != port) {
        throw new Error('invalid port in candidate line: ' + candidate);
    }
    return {
        address: address,
        port: port
    };
};
exports.setCandidateLineEndpoint = function (candidate, endpoint) {
    var lines = splitHostCandidateLine_(candidate);
    lines[4] = endpoint.address;
    lines[5] = endpoint.port.toString();
    return lines.join(' ');
};
// This function implements a heuristic to select the single candidate
// that is most likely to work for this connection.  The heuristic
// expresses a preference ordering:
// Most preferred: public IP address bound as a host candidate
//  - rare outside of servers, but offers the very best connectivity
// Next best: server-reflexive IP address
//  - most common
// Worst: private IP address in a host candidate
//  - indicates that STUN has failed.  Connection is still possible
//    if the other side is directly routable.
// If none of these are present, the function will throw an exception.
// TODO: Allow selecting more than one public address.  This would help
// when there are multiple interfaces or IPv6 and IPv4.
exports.selectPublicAddress = function (candidates) {
    // TODO: Note that we cannot currently support IPv6 addresses:
    //         https://github.com/uProxy/uproxy/issues/1107
    var publicHostCandidates = [];
    var srflxCandidates = [];
    var privateHostCandidates = [];
    for (var i = 0; i < candidates.length; ++i) {
        var line = candidates[i].candidate;
        var tokens = splitCandidateLine_(line);
        if (tokens[2].toLowerCase() != 'udp') {
            continue;
        }
        var typ = tokens[7];
        if (typ === 'srflx') {
            var srflxAddress = tokens[4];
            if (ipaddr.process(srflxAddress).kind() === 'ipv6') {
                continue;
            }
            var port = parseInt(tokens[5]);
            if (tokens[8] != 'raddr') {
                throw new Error('no raddr in candidate line: ' + line);
            }
            var raddr = tokens[9];
            if (ipaddr.process(raddr).kind() === 'ipv6') {
                continue;
            }
            if (tokens[10] != 'rport') {
                throw new Error('no rport in candidate line: ' + line);
            }
            var rport = parseInt(tokens[11]);
            srflxCandidates.push({
                external: {
                    address: srflxAddress,
                    port: port
                },
                internal: {
                    address: raddr,
                    port: rport
                }
            });
        }
        else if (typ === 'host') {
            var hostAddress = ipaddr.process(tokens[4]);
            // Store the host address in case no srflx candidates are found.
            if (hostAddress.kind() !== 'ipv6') {
                var endpoint = {
                    address: tokens[4],
                    port: parseInt(tokens[5])
                };
                if (hostAddress.range() === 'unicast') {
                    publicHostCandidates.push(endpoint);
                }
                else {
                    privateHostCandidates.push(endpoint);
                }
            }
        }
    }
    if (publicHostCandidates.length > 0) {
        return {
            internal: publicHostCandidates[0],
            external: publicHostCandidates[0]
        };
    }
    else if (srflxCandidates.length > 0) {
        return srflxCandidates[0];
    }
    else if (privateHostCandidates.length > 0) {
        return {
            internal: privateHostCandidates[0],
            external: privateHostCandidates[0]
        };
    }
    throw new Error('no srflx or host candidate found');
};
// Returns a key for use with mirrorPipes_.
var makeEndpointKey_ = function (endpoint) {
    return endpoint.address + ':' + endpoint.port;
};
// Retry an async function with exponential backoff for up to 2 seconds
// before failing.
var retry_ = function (func, delayMs) {
    delayMs = delayMs || 10;
    return func().catch(function (err) {
        delayMs *= 2;
        if (delayMs > 2000) {
            return Promise.reject(err);
        }
        return new Promise(function (F, R) {
            setTimeout(function () {
                _this.retry_(func, delayMs).then(F, R);
            }, delayMs);
        });
    });
};
/**
 * A uproxypeerconnection-like Freedom module which establishes obfuscated
 * connections.
 *
 * DTLS packets are intercepted by pointing WebRTC at a local "forwarding"
 * port; connectivity to the remote host is achieved with the help of
 * another preceding, short-lived, peer-to-peer connection.
 *
 * This is mostly a thin wrapper over uproxypeerconnection except for the
 * magic required during setup.
 *
 * TODO: Give the uproxypeerconnections name, to help debugging.
 * TODO: Allow obfuscation parameters be configured.
 */
var Connection = (function () {
    function Connection(probeRtcPc, peerName) {
        var _this = this;
        // The list of all candidates returned by the probe connection.
        this.probeCandidates_ = [];
        this.onceProbingComplete_ = new Promise(function (F, R) {
            _this.probingComplete_ = F;
        });
        this.onceHaveWebRtcEndpoint_ = new Promise(function (F, R) {
            _this.haveWebRtcEndpoint_ = F;
        });
        this.onceHaveRemoteEndpoint_ = new Promise(function (F, R) {
            _this.haveRemoteEndpoint_ = F;
        });
        this.onceHaveForwardingSocketEndpoint_ = new Promise(function (F, R) {
            _this.haveForwardingSocketEndpoint_ = F;
        });
        // A map from remote transport addresses to local pipes that represent them.
        this.mirrorPipes_ = {};
        this.configureProbeConnection_ = function (freedomPc) {
            var probePeerName = _this.peerName + '-probe';
            _this.probeConnection_ = new peerconnection.PeerConnectionClass(freedomPc, probePeerName);
            _this.probeConnection_.signalForPeerQueue.setSyncHandler(function (message) {
                if (message.type === 2 /* CANDIDATE */) {
                    _this.probeCandidates_.push(message.candidate);
                }
                else if (message.type === 3 /* NO_MORE_CANDIDATES */) {
                    _this.probeConnection_.close().then(function () {
                        _this.probingComplete_(exports.selectPublicAddress(_this.probeCandidates_));
                    });
                }
            });
            _this.probeConnection_.negotiateConnection();
        };
        // Add and return a local pipe that represents a specific remote address
        // for both send and receive.
        this.addLocalPipe_ = function (webRtcEndpoint, remoteEndpoint, publicPipe) {
            log.info('%1: Adding local pipe between %2 and %3', _this.peerName, webRtcEndpoint, remoteEndpoint);
            var key = makeEndpointKey_(remoteEndpoint);
            if (_this.mirrorPipes_[key]) {
                log.warn('%1: Got redundant call to add local pipe for %2', _this.peerName, key);
                // Return the pipe, but wait until it's ready.
                return _this.mirrorPipes_[key].getLocalEndpoint().then(function (ignored) {
                    return _this.mirrorPipes_[key];
                });
            }
            var localPipe = freedom['churnPipe']();
            _this.mirrorPipes_[key] = localPipe;
            // Packets received by this pipe should be obfuscated and forwarded
            // to the corresponding remote endpoint.
            localPipe.on('message', function (m) {
                publicPipe.sendTo(m.data, remoteEndpoint);
            });
            return localPipe.bind('127.0.0.1', 0, webRtcEndpoint.address, webRtcEndpoint.port, 'none', undefined, undefined).catch(function (e) {
                log.error('%1: error establishing local pipe: %2', _this.peerName, e.message);
            }).then(localPipe.getLocalEndpoint).then(function (forwardingSocketEndpoint) {
                log.info('%1: configured local pipe between %2 and %3', _this.peerName, JSON.stringify(forwardingSocketEndpoint), JSON.stringify(webRtcEndpoint));
                return localPipe;
            });
        };
        this.bindPublicPipe_ = function (publicPipe, local, remote) {
            return publicPipe.bind(local.address, local.port, remote.address, remote.port, 'caesar', new Uint8Array([13]).buffer, '{}');
        };
        // Establishes the two pipes required to sustain the obfuscated
        // connection:
        //  - a non-obfuscated, local only, between WebRTC and a new,
        //    automatically allocated, port
        //  - remote, obfuscated, port
        this.configurePipes_ = function (webRtcEndpoint, remoteEndpoint, natEndpoints) {
            log.debug('%1: configuring pipes...', _this.peerName);
            var publicPipe = freedom['churnPipe']();
            // This retry is needed because the browser releases the UDP port
            // asynchronously after we call close() on the RTCPeerConnection, so
            // this.bindPublicPipe_ may initially fail, until the port is released.
            retry_(function () {
                return _this.bindPublicPipe_(publicPipe, natEndpoints.internal, remoteEndpoint);
            }).then(function () {
                log.info('%1: configured obfuscating pipe between %2 and %3', _this.peerName, JSON.stringify(natEndpoints.internal), JSON.stringify(remoteEndpoint));
            }, function (e) {
                log.error('%1: error establishing public pipe between %2 and %3: %4', _this.peerName, makeEndpointKey_(natEndpoints.internal), makeEndpointKey_(remoteEndpoint), e.message);
            });
            publicPipe.on('message', function (m) {
                // This is the particular local pipe associated with this sender.
                var localPipe = _this.mirrorPipes_[makeEndpointKey_(m.source)];
                if (localPipe) {
                    // Note: due to asynchronous setup, it's possible that this pipe
                    // has not yet been bound.  Hopefully, the send call will be
                    // queued behind the bind call.  If not, the packet may just be
                    // dropped (which should be acceptable for a brief period).
                    localPipe.send(m.data);
                }
                else if (_this.pcState == 0 /* WAITING */ || _this.pcState == 1 /* CONNECTING */) {
                    log.info('%1: Got packet from new source; peer has symmetric NAT?', _this.peerName);
                    _this.addLocalPipe_(webRtcEndpoint, m.source, publicPipe).then(function (localPipe) {
                        // Don't drop the first packet.
                        localPipe.send(m.data);
                    });
                }
                else {
                    log.warn('%1: Received unexpected packet of length %2 from %3' + ' while in state %4', _this.peerName, m.data.byteLength, makeEndpointKey_(m.source), _this.pcState);
                }
            });
            _this.addLocalPipe_(webRtcEndpoint, remoteEndpoint, publicPipe).then(function (pipe) {
                return pipe.getLocalEndpoint();
            }).then(_this.haveForwardingSocketEndpoint_);
        };
        this.configureObfuscatedConnection_ = function () {
            // We use an empty configuration to ensure that no STUN servers are pinged.
            var obfConfig = {
                iceServers: []
            };
            var obfPeerName = _this.peerName + '-obfuscated';
            var freedomPc = freedom['core.rtcpeerconnection'](obfConfig);
            _this.obfuscatedConnection_ = new peerconnection.PeerConnectionClass(freedomPc, obfPeerName);
            _this.obfuscatedConnection_.signalForPeerQueue.setSyncHandler(function (message) {
                // Super-paranoid check: remove candidates from SDP messages.
                // This can happen if a connection is re-negotiated.
                // TODO: We can safely remove this once we can reliably interrogate
                //       peerconnection endpoints.
                if (message.type === 0 /* OFFER */ || message.type === 1 /* ANSWER */) {
                    message.description.sdp = exports.filterCandidatesFromSdp(message.description.sdp);
                }
                if (message.type === 2 /* CANDIDATE */) {
                    try {
                        if (!message.candidate || !message.candidate.candidate) {
                            throw new Error('no candidate line');
                        }
                        var address = exports.extractEndpointFromCandidateLine(message.candidate.candidate);
                        // TODO: We cannot currently support IPv6 addresses:
                        //         https://github.com/uProxy/uproxy/issues/1107
                        if (ipaddr.process(address.address).kind() === 'ipv6') {
                            throw new Error('ipv6 unsupported');
                        }
                        _this.haveWebRtcEndpoint_(address);
                        message.candidate.candidate = exports.setCandidateLineEndpoint(message.candidate.candidate, {
                            address: '0.0.0.0',
                            port: 0
                        });
                    }
                    catch (e) {
                        log.debug('%1: ignoring candidate line %2: %3', _this.peerName, JSON.stringify(message), e.message);
                    }
                }
                var churnSignal = {
                    webrtcMessage: message
                };
                _this.signalForPeerQueue.handle(churnSignal);
            });
            // NOTE: Replacing |this.dataChannels| in this way breaks recursive nesting.
            // If the caller or |obfuscatedConnection_| applies the same approach,
            // the code will break in hard-to-debug fashion.  This could be
            // addressed by using a javascript "getter", or by changing the
            // peerconnection.PeerConnection API.
            _this.dataChannels = _this.obfuscatedConnection_.dataChannels;
            _this.peerOpenedChannelQueue = _this.obfuscatedConnection_.peerOpenedChannelQueue;
        };
        this.negotiateConnection = function () {
            return _this.obfuscatedConnection_.negotiateConnection();
        };
        // Forward the message to the relevant stage: churn-pipe or obfuscated.
        // In the case of obfuscated signalling channel messages, we inject our
        // local forwarding socket's endpoint.
        this.handleSignalMessage = function (churnMessage) {
            if (churnMessage.publicEndpoint !== undefined) {
                _this.haveRemoteEndpoint_(churnMessage.publicEndpoint);
            }
            if (churnMessage.webrtcMessage) {
                var message = churnMessage.webrtcMessage;
                if (message.type === 2 /* CANDIDATE */) {
                    _this.onceHaveForwardingSocketEndpoint_.then(function (forwardingSocketEndpoint) {
                        message.candidate.candidate = exports.setCandidateLineEndpoint(message.candidate.candidate, forwardingSocketEndpoint);
                        _this.obfuscatedConnection_.handleSignalMessage(message);
                    });
                }
                else if (message.type == 0 /* OFFER */ || message.type == 1 /* ANSWER */) {
                    // Remove candidates from the SDP.  This is redundant, but ensures
                    // that a bug in the remote client won't cause us to send
                    // unobfuscated traffic.
                    message.description.sdp = exports.filterCandidatesFromSdp(message.description.sdp);
                    _this.obfuscatedConnection_.handleSignalMessage(message);
                }
            }
        };
        this.openDataChannel = function (channelLabel, options) {
            return _this.obfuscatedConnection_.openDataChannel(channelLabel);
        };
        this.close = function () {
            return _this.obfuscatedConnection_.close();
        };
        this.toString = function () {
            return _this.obfuscatedConnection_.toString();
        };
        this.peerName = peerName || 'churn-connection-' + (++Connection.internalConnectionId_);
        this.signalForPeerQueue = new handler.Queue();
        // Configure the probe connection.  Once it completes, inform the remote
        // peer which public endpoint we will be using.
        this.onceProbingComplete_.then(function (endpoints) {
            _this.signalForPeerQueue.handle({
                publicEndpoint: endpoints.external
            });
        });
        // Start the obfuscated connection.
        this.configureObfuscatedConnection_();
        // Once the obfuscated connection's local endpoint is known, the remote
        // peer has sent us its public endpoint, and probing is complete, we can
        // configure the obfuscating pipe and allow traffic to flow.
        this.configureProbeConnection_(probeRtcPc);
        Promise.all([this.onceHaveWebRtcEndpoint_, this.onceHaveRemoteEndpoint_, this.onceProbingComplete_]).then(function (answers) {
            _this.configurePipes_(answers[0], answers[1], answers[2]);
        });
        // Handle |pcState| and related promises.
        this.pcState = 0 /* WAITING */;
        this.onceConnecting = this.obfuscatedConnection_.onceConnecting.then(function () {
            _this.pcState = 1 /* CONNECTING */;
        });
        this.onceConnected = this.obfuscatedConnection_.onceConnected.then(function () {
            _this.pcState = 2 /* CONNECTED */;
        });
        this.onceDisconnected = this.obfuscatedConnection_.onceDisconnected.then(function () {
            _this.pcState = 3 /* DISCONNECTED */;
        });
        // Debugging.
        this.onceProbingComplete_.then(function (endpoint) {
            log.debug('%1: NAT endpoints of probe connection are %2', _this.peerName, JSON.stringify(endpoint));
        });
        this.onceHaveWebRtcEndpoint_.then(function (endpoint) {
            log.debug('%1: obfuscated connection is bound to %2', _this.peerName, JSON.stringify(endpoint));
        });
        this.onceHaveRemoteEndpoint_.then(function (endpoint) {
            log.debug('%1: remote peer is contactable at %2', _this.peerName, JSON.stringify(endpoint));
        });
    }
    Connection.internalConnectionId_ = 0;
    return Connection;
})();
exports.Connection = Connection;
