/// <reference path='../../../third_party/typings/es6-promise/es6-promise.d.ts' />
/// <reference path='../../../third_party/freedom-typings/udp-socket.d.ts' />
var logging = require('../../../third_party/uproxy-lib/logging/logging');
var log = new logging.Log('udprelay');
/**
 * A UDP-based "relay" server intended for use as part of a SOCKS5 proxy:
 *   http://www.ietf.org/rfc/rfc1928.txt
 *
 * Briefly, here's how to use this:
 *  - create an instance of this class
 *  - call bind (you probably want to specify port zero to have the system
 *    pick a free port)
 *  - call getInfo, to discover on which port the relay is listening
 *  - (the caller can now return the address and port back to the SOCKS
 *    client)
 *  - call setDataReceivedHandler
 *  - this class will invoke the setDataReceivedHandler each time a datagram
 *    is received on the socket; the full message is sent, including SOCKS
 *    UDP headers
 *  - (the caller can relay the message across the datachannel)
 *  - call sendRemoteReply for each datagram received from remote hosts; this
 *    will be sent to the client, and should include the same SOCKS UDP
 *    header received in the original request
 *  - call destroy to clean up, typically when the TCP connection on which
 *    the UDP_ASSOCIATE was negotiated is terminated
 *
 * One relay should be created in response to each UDP_ASSOCIATE command.
 *
 * Other notes:
 *  - while the RFC states that the relay MUST drop any message originating
 *    from an IP other than that which requested the association, this
 *    implementation makes no effort to do so (this isn't urgent because we
 *    typically only listen locally)
 *  - similarly, we make no effort to respect the DST.PORT and DST.ADDR fields
 *    specified by the client during the handshake: having run various proxy
 *    clients it seems that these are rarely specified anyway (which is fine
 *    according to section 6 of the RFC)
 *    and, in any case, we are typically only listening locally
 *  - we make no attempt to implement fragmentation (see section 7 of the
 *    RFC)
 *
 * TODO: this is so similar to udprelay.ts that they can almost certainly
 *       be merged into one
 */
var UdpRelay = (function () {
    function UdpRelay() {
        var _this = this;
        /**
         * Listens for onData events.
         * The socket must be bound.
         */
        this.attachSocketHandler_ = function () {
            _this.socket_.on('onData', _this.onSocksClientData_);
        };
        this.onSocksClientData_ = function (recvFromInfo) {
            // Record the host:port from which the client is sending us datagrams.
            // This is where we'll relay any replies from remote servers.
            // TODO: check if these change over the liftime of the relay
            _this.clientAddress_ = recvFromInfo.address;
            _this.clientPort_ = recvFromInfo.port;
            if (_this.dataReceivedHandler) {
                _this.dataReceivedHandler(recvFromInfo.data);
            }
        };
        // TODO(yangoon): add destroy() method
        /**
         * Returns the address on which the local socket associated with this
         * relay is listening.
         */
        this.getAddress = function () {
            return _this.address_;
        };
        /**
         * Returns the port on which the local socket associated with this
         * relay is listening.
         */
        this.getPort = function () {
            return _this.port_;
        };
        this.socket_ = freedom['core.udpsocket']();
    }
    /**
     * Returns a promise to create a socket, bind to the specified address and
     * port, and start relaying events. Specify port zero to have the system
     * choose a free port.
     */
    UdpRelay.prototype.bind = function (address, port) {
        var _this = this;
        return this.socket_.bind(address, port).then(function (resultCode) {
            // Ensure the listen was successful.
            if (resultCode != 0) {
                return Promise.reject(new Error('listen failed on ' + _this.address_ + ':' + _this.port_ + ' with result code ' + resultCode));
            }
            return Promise.resolve(resultCode);
        }).then(this.socket_.getInfo).then(function (socketInfo) {
            // Record the address and port on which our socket is listening.
            _this.address_ = socketInfo.localAddress;
            _this.port_ = socketInfo.localPort;
            log.info('listening on ' + _this.address_ + ':' + _this.port_);
        }).then(this.attachSocketHandler_);
    };
    /**
     * Sets the function to be called when data is received from the client.
     * This is intended for relaying datagrams from the client across the
     * datachannel from a remote server. The full datagram as received from the
     * client is sent, complete with SOCKS headers.
     */
    UdpRelay.prototype.setDataReceivedHandler = function (callback) {
        this.dataReceivedHandler = callback;
    };
    /**
     * Returns a promise to send data to the client.
     * This is intended for relaying responses from remote servers back to
     * the client.
     */
    UdpRelay.prototype.sendRemoteReply = function (buffer) {
        if (!this.clientAddress_) {
            throw new Error('cannot send data to client before it sends data');
        }
        return this.socket_.sendTo(buffer, this.clientAddress_, this.clientPort_);
    };
    return UdpRelay;
})();
module.exports = UdpRelay;
