/// <reference path='../../../third_party/typings/es6-promise/es6-promise.d.ts' />
var handler = require('../../../third_party/uproxy-lib/handler/queue');
var queue = require('../../../third_party/uproxy-lib/queue/queue');
var logging = require('../../../third_party/uproxy-lib/logging/logging');
var log = new logging.Log('pool');
// This is the only exported class in this module.  It mimics the data channel
// aspects of the PeerConnection interface.  Internally, it provides a pool
// of channels that keeps old channels for reuse instead of closing them, and
// makes new channels as needed when the pool runs dry.  Crucially, the local
// and remote pools do not interfere with each other, even if they use the
// same labels, because the browser ensures that data channels created by each
// peer are drawn from separate ID spaces (odd vs. even).
var Pool = (function () {
    function Pool(pc, name_) {
        var _this = this;
        this.openDataChannel = function () {
            return _this.localPool_.openDataChannel();
        };
        this.localPool_ = new LocalPool(pc, name_);
        var remotePool = new RemotePool(pc, name_);
        this.peerOpenedChannelQueue = remotePool.peerOpenedChannelQueue;
    }
    return Pool;
})();
// Manages a pool of data channels opened by this peer.  The only public method
// is openDataChannel.
var LocalPool = (function () {
    function LocalPool(pc_, name_) {
        var _this = this;
        this.pc_ = pc_;
        this.name_ = name_;
        this.numChannels_ = 0;
        // Channels which have been closed, and may be re-opened.
        this.pool_ = new queue.Queue();
        this.openDataChannel = function () {
            return _this.reuseOrCreate_().then(function (channel) {
                return channel.open().then(function () {
                    // When this channel closes, reset it and return it to the pool.
                    channel.onceClosed.then(function () {
                        _this.onChannelClosed_(channel);
                    });
                    return channel;
                });
            });
        };
        this.reuseOrCreate_ = function () {
            // If there are no channels available right now, open a new one.
            // TODO: limit the number of channels (probably should be <=256).
            if (_this.pool_.length > 0) {
                var channel = _this.pool_.shift();
                log.debug('%1: channel requested, pulled %2 from pool (%3 remaining)', _this.name_, channel.getLabel(), _this.pool_.length);
                return Promise.resolve(channel);
            }
            else {
                log.debug('%1: channel requested, creating new', _this.name_);
                return _this.openNewChannel_();
            }
        };
        // Creates and returns a new channel, wrapping it.
        this.openNewChannel_ = function () {
            return _this.pc_.openDataChannel('p' + _this.numChannels_++).then(function (dc) {
                return dc.onceOpened.then(function () {
                    return new PoolChannel(dc);
                });
            });
        };
        // Resets the channel, making it ready for use again, and adds it
        // to the pool.
        this.onChannelClosed_ = function (poolChannel) {
            if (!poolChannel.reset()) {
                return;
            }
            _this.pool_.push(poolChannel);
            log.debug('%1: returned channel %2 to the pool (new size: %3)', _this.name_, poolChannel.getLabel(), _this.pool_.length);
        };
    }
    return LocalPool;
})();
// Tracks a pool of channels that were opened by the remote peer.
var RemotePool = (function () {
    function RemotePool(pc_, name_) {
        var _this = this;
        this.pc_ = pc_;
        this.name_ = name_;
        this.peerOpenedChannelQueue = new handler.Queue();
        this.onNewChannel_ = function (dc) {
            log.debug('%1: remote side created new channel: %2', _this.name_, dc.getLabel());
            dc.onceOpened.then(function () {
                var poolChannel = new PoolChannel(dc);
                _this.listenForOpenAndClose_(poolChannel);
            });
        };
        this.listenForOpenAndClose_ = function (poolChannel) {
            poolChannel.onceOpened.then(function () {
                _this.peerOpenedChannelQueue.handle(poolChannel);
            });
            poolChannel.onceClosed.then(function () {
                if (!poolChannel.reset()) {
                    return;
                }
                _this.listenForOpenAndClose_(poolChannel);
            });
        };
        this.pc_.peerOpenedChannelQueue.setSyncHandler(this.onNewChannel_);
    }
    return RemotePool;
})();
// These are the two control messages used.  To help debugging, and
// improve forward-compatibility, we send the string name on the wire,
// not the numerical enum value.  Therefore, these names are part of
// the normative protocol, and will break compatibility if changed.
var ControlMessage;
(function (ControlMessage) {
    ControlMessage[ControlMessage["OPEN"] = 0] = "OPEN";
    ControlMessage[ControlMessage["CLOSE"] = 1] = "CLOSE";
})(ControlMessage || (ControlMessage = {}));
var State;
(function (State) {
    State[State["OPEN"] = 0] = "OPEN";
    State[State["CLOSING"] = 1] = "CLOSING";
    State[State["CLOSED"] = 2] = "CLOSED";
    State[State["PERMANENTLY_CLOSED"] = 3] = "PERMANENTLY_CLOSED";
})(State || (State = {}));
// Each PoolChannel wraps an actual DataChannel, and provides behavior
// that is intended to be indistinguishable to the caller.  However,
// close() does not actually close the underlying channel.  Instead,
// it sends an in-band control message indicating the close, and the
// channel is returned to the pool of inactive channels, ready for
// reuse when the client asks for a new channel.
var PoolChannel = (function () {
    // dc_.onceOpened must already have resolved
    function PoolChannel(dc_) {
        var _this = this;
        this.dc_ = dc_;
        this.state_ = 2 /* CLOSED */;
        this.reset = function () {
            if (_this.state_ !== 2 /* CLOSED */) {
                return false;
            }
            _this.dataFromPeerQueue = new handler.Queue();
            _this.lastDataFromPeerHandled_ = Promise.resolve();
            _this.onceOpened = new Promise(function (F, R) {
                _this.fulfillOpened_ = F;
            });
            _this.onceClosed = new Promise(function (F, R) {
                _this.fulfillClosed_ = F;
            });
            _this.onceOpened.then(function () {
                if (_this.state_ === 2 /* CLOSED */) {
                    _this.state_ = 0 /* OPEN */;
                }
            });
            _this.onceClosed.then(function () {
                if (_this.state_ !== 3 /* PERMANENTLY_CLOSED */) {
                    _this.state_ = 2 /* CLOSED */;
                }
            });
            return true;
        };
        this.getLabel = function () {
            return _this.dc_.getLabel();
        };
        this.send = function (data) {
            if (_this.state_ !== 0 /* OPEN */) {
                return Promise.reject(new Error('Can\'t send while closed'));
            }
            // To distinguish control messages from application data, all string
            // messages are encapsulated in a JSON layer. Binary messages are unaffected.
            if (data.str) {
                return _this.dc_.send({
                    str: JSON.stringify({
                        data: data.str
                    })
                });
            }
            return _this.dc_.send(data);
        };
        this.sendControlMessage_ = function (controlMessage) {
            log.debug('%1: sending control message: %2', _this.getLabel(), ControlMessage[controlMessage]);
            return _this.dc_.send({
                str: JSON.stringify({
                    control: ControlMessage[controlMessage]
                })
            });
        };
        this.onDataFromPeer_ = function (data) {
            if (data.str) {
                try {
                    var msg = JSON.parse(data.str);
                }
                catch (e) {
                    log.error('%1: Got non-JSON string: %2', _this.getLabel(), data.str);
                    return;
                }
                if (typeof msg.data === 'string') {
                    _this.lastDataFromPeerHandled_ = _this.dataFromPeerQueue.handle({ str: msg.data });
                }
                else if (typeof msg.control === 'string') {
                    _this.onControlMessage_(msg.control);
                }
                else {
                    log.error('No data or control message found');
                }
                return;
            }
            _this.lastDataFromPeerHandled_ = _this.dataFromPeerQueue.handle(data);
        };
        this.onControlMessage_ = function (controlMessage) {
            log.debug('%1: received control message: %2', _this.getLabel(), controlMessage);
            if (controlMessage === ControlMessage[0 /* OPEN */]) {
                if (_this.state_ === 0 /* OPEN */) {
                    log.warn('%1: Got redundant open message', _this.getLabel());
                }
                _this.fulfillOpened_();
            }
            else if (controlMessage === ControlMessage[1 /* CLOSE */]) {
                if (_this.state_ === 0 /* OPEN */) {
                    _this.state_ = 1 /* CLOSING */;
                    // Drain messages, then ack the close.
                    _this.lastDataFromPeerHandled_.then(function () {
                        return _this.sendControlMessage_(1 /* CLOSE */);
                    }).then(_this.fulfillClosed_);
                }
                else if (_this.state_ === 1 /* CLOSING */) {
                    // We both sent a "close" command at the same time.
                    _this.fulfillClosed_();
                }
                else if (_this.state_ === 2 /* CLOSED */) {
                    log.warn('%1: Got redundant close message', _this.getLabel());
                }
            }
            else {
                log.error('%1: unknown control message: %2', _this.getLabel(), controlMessage);
            }
        };
        this.getBrowserBufferedAmount = function () {
            return _this.dc_.getBrowserBufferedAmount();
        };
        this.getJavascriptBufferedAmount = function () {
            return _this.dc_.getJavascriptBufferedAmount();
        };
        this.isInOverflow = function () {
            return _this.dc_.isInOverflow();
        };
        this.setOverflowListener = function (listener) {
            _this.dc_.setOverflowListener(listener);
        };
        // New method for PoolChannel, not present in the DataChannel interface.
        this.open = function () {
            log.debug(_this.getLabel() + ': open');
            if (_this.state_ === 0 /* OPEN */) {
                return Promise.reject(new Error('channel is already open'));
            }
            _this.sendControlMessage_(0 /* OPEN */);
            // Immediate open; there is no open-ack
            _this.fulfillOpened_();
            return _this.onceOpened;
        };
        this.close = function () {
            log.debug('%1: close', _this.getLabel());
            if (_this.state_ !== 0 /* OPEN */) {
                return;
            }
            _this.state_ = 1 /* CLOSING */;
            _this.sendControlMessage_(1 /* CLOSE */);
            return _this.onceClosed;
        };
        this.toString = function () {
            return 'PoolChannel(' + _this.dc_.toString() + ')';
        };
        this.reset();
        this.dc_.dataFromPeerQueue.setSyncHandler(this.onDataFromPeer_);
        this.dc_.onceClosed.then(function () {
            _this.state_ = 3 /* PERMANENTLY_CLOSED */;
            _this.fulfillClosed_();
        });
    }
    return PoolChannel;
})();
module.exports = Pool;
